\Module{Configuration}

[Discuss the file format and rationale.]

|#include <stdlib.h>
|#include <stdio.h>
|#include <string.h>
|#include <ctype.h>
|
|#include "neural.h"
|#include "protos.h"



Function prototypes local to this module.

|static char *SkipBlanks           (char *s);
|static char *SkipNonBlanks        (char *s);
|static bool ReadLine              (char *s, int slen, FILE *f);
|static bool ReadRandSeed          (char *s, FILE *f);
|static bool ReadNumNeurons        (char *s, FILE *f);
|static bool ReadNumInputs         (char *s, FILE *f);
|static bool ReadNumOutputs        (char *s, FILE *f);
|static bool ReadPloidy            (char *s, FILE *f);
|static bool ReadNumDemes          (char *s, FILE *f);
|static bool ReadNumOrganisms      (char *s, FILE *f);
|static bool ReadNumGenerations    (char *s, FILE *f);
|static bool ReadNumMutations      (char *s, FILE *f);
|static bool ReadNumIterations     (char *s, FILE *f);
|static bool ReadInitZeroFlag      (char *s, FILE *f);
|static bool ReadDiffEqFlag        (char *s, FILE *f);
|static bool ReadNormalizeAllFlag  (char *s, FILE *f);
|static bool ReadErrorWeightAlpha  (char *s, FILE *f);
|static bool ReadErrorWeightBeta   (char *s, FILE *f);
|static bool ReadVarMatrix         (char *s, FILE *f);
|static void LoadDefaultVarMatrix  (void);
|static bool ReadTrainSet          (char *s, FILE *f);
|static bool ReadTestSet           (char *s, FILE *f);
|static bool ReadTrainTestSet      (char *s, bool train);
|static real ExtractReal           (char **s);
|static bool CheckRange            (int n, int nmin, int nmax, char *name);
|static bool CheckMax              (int n, int nmax, char *name);
|static bool CheckDefined          (int n, char *name);


All configuration data is held in a globally accessible structure in order to
avoid passing the address of the structure among and between dozens of
functions.  Early versions of the program allocated this data structure on
the stack of a top-level function and passed the pointer everywhere (it was a
mess).  Please refer to the header file for the layout of the data structure.

|Config c;


\Section{Limits}

   These limits are based on expected maximum memory availability, but may need
to change from system to system.


Neuron limits.

|#define  MIN_NEURONS  2
|#define  MAX_NEURONS  50
|
|#define  MIN_INPUT    1
|#define  MAX_INPUT    20
|
|#define  MIN_HIDDEN   0
|#define  MAX_HIDDEN   50
|
|#define  MIN_OUTPUT   1
|#define  MAX_OUTPUT   20


Ploidy limits.

|#define  MIN_PLOIDY   1
|#define  MAX_PLOIDY   10


Deme and population limits.

|#define  MIN_DEMES    1
|#define  MAX_DEMES    100
|
|#define  MIN_POP      1
|#define  MAX_POP      1000


Train and test set limits.

|#define  MAX_TRAIN    300
|#define  MAX_TEST     300


A special undefined value is used while reading the configuration file to
ensure that certain values are defined before defining others.

|#define  UNDEFINED  (-1)


\Section{Read Configuration from File}

   This routine reads the configuration file, relying upon a collection of
lower-level routines to read the various parts.  A table of keywords is scanned
for each source input line of the file.  When a match is found, the
corresponding subroutine is called to process the input line.

\ENTRY: 'filename' specifies a configuration file to read.

\EXIT:  'ReadConfiguration' returns TRUE if successful, FALSE if unsuccessful.

|bool ReadConfiguration (char *filename)
|{
|  static struct
|  {
|    char  *kname;                           `Expected keyword in file.
|    int   klen;                             `Keyword length (filled in below).
|    bool  (*kread)(char *, FILE *);         `Function to extract data.
|  }
|  keys[] =
|  {
|    { "RandSeed",     0,  &ReadRandSeed         },
|    { "Neurons",      0,  &ReadNumNeurons       },
|    { "Input",        0,  &ReadNumInputs        },
|    { "Output",       0,  &ReadNumOutputs       },
|    { "Ploidy",       0,  &ReadPloidy           },
|    { "Demes",        0,  &ReadNumDemes         },
|    { "Organisms",    0,  &ReadNumOrganisms     },
|    { "Population",   0,  &ReadNumOrganisms     },
|    { "Generations",  0,  &ReadNumGenerations   },
|    { "Mutations",    0,  &ReadNumMutations     },
|    { "Iterations",   0,  &ReadNumIterations    },
|    { "InitZero",     0,  &ReadInitZeroFlag     },
|    { "DiffEq",       0,  &ReadDiffEqFlag       },
|    { "NormalizeAll", 0,  &ReadNormalizeAllFlag },
|    { "SynapseAlpha", 0,  &ReadErrorWeightAlpha },
|    { "SynapseBeta",  0,  &ReadErrorWeightBeta  },
|    { "Matrix",       0,  &ReadVarMatrix        },
|    { "Train",        0,  &ReadTrainSet         },
|    { "Test",         0,  &ReadTestSet          },
|  };
|
|  FILE *f;
|  int k;
|  bool ret = TRUE;

Initialize key lengths.

|  for (k = 0; k < NUM_ELEMENTS(keys); k++)
|    keys[k].klen = strlen(keys[k].kname);

Initialize configuration structure.

|  c.ni = UNDEFINED;
|  c.nh = UNDEFINED;
|  c.no = UNDEFINED;
|  c.nn = UNDEFINED;
|
|  c.seed   = 0;
|  c.ploidy = 1;
|  c.ndemes = 1;
|  c.npop   = 1;
|  c.ngen   = 0;
|  c.nmut   = 1;
|  c.niter  = 2;
|
|  c.ntrain = 0;
|  c.ntest  = 0;
|
|  c.train  = (Vector **) CreateArray1(MAX_TRAIN, sizeof(Vector *));
|  c.test   = (Vector **) CreateArray1(MAX_TEST,  sizeof(Vector *));
|
|  c.initzero    = FALSE;
|  c.diffeq      = FALSE;
|  c.normall     = FALSE;
|
|  c.errorweight = FALSE;
|  c.walpha      = (real) 1.0;
|  c.wbeta       = (real) 1.0;
|
|  c.M = (Matrix *) 0;
|  c.B = (Vector *) 0;

Read the configuration file.

|#define strncasecmp strncmp

|  if (f = fopen(filename, "rt"))
|  {
|    char line[256];
|
|    printf("Reading data file \"%s\"...\n", filename);
|
|    while (ReadLine(line, sizeof(line), f))
|      for (k = 0; k < NUM_ELEMENTS(keys); k++)
-        if ((strnicmp(line, keys[k].kname, keys[k].klen) == 0) &&
|        if ((strncasecmp(line, keys[k].kname, keys[k].klen) == 0) &&
|            (isspace(line[keys[k].klen]) || !line[keys[k].klen]))
|          ret &= keys[k].kread(line + keys[k].klen, f);
|
|    if (!c.M || !c.B)
|      LoadDefaultVarMatrix();
|
|    fclose(f);
|    return ret;
|  }
|
|  else
|  {
|    printf("Unable to open data file %s.\n", filename);
|    return FALSE;
|  }
|}



\Section{Read Configuration from Argument List}

   Selected configuration file setting may be overriden by calling this
function with the command line arguments as supplied to 'main'.  The number of
input, output, and hidden-layer neurons is fixed when reading the configuration
file, since matrices are allocated at that time.  Moreover, the variable matrix
could not be specified on the command line anyway.

   The following values can be changed on the command line:

\item{}   RandSeed
\item{}   Ploidy
\item{}   Demes
\item{}   Population
\item{}   Generations
\item{}   Mutations
\item{}   Iterations
\item{}   InitZero
\item{}   WeightAlpha
\item{}   WeightBeta

Please refer to the table in the code below for the associated command-line
switches.

\ENTRY: 'argc' contains the number of command-line arguments.\n
        'argv' points to an array of command-line arguments.

\EXIT:  'ReadArgConfiguration' returns 'TRUE' if successful, 'FALSE' if
         unsuccessful.

|bool ReadArgConfiguration (int argc, char *argv[])
|{
|  static struct
|  {
|    char  *kname;                           `Expected command-line switch.
|    int   klen;                             `Keyword length (filled in below).
|    bool  (*kread)(char *, FILE *);         `Function to extract data.
|  }
|  keys[] =
|  {
|    { "-r",  0,  &ReadRandSeed         },
|    { "-p",  0,  &ReadPloidy           },
|    { "-d",  0,  &ReadNumDemes         },
|    { "-o",  0,  &ReadNumOrganisms     },
|    { "-g",  0,  &ReadNumGenerations   },
|    { "-m",  0,  &ReadNumMutations     },
|    { "-i",  0,  &ReadNumIterations    },
|    { "-z",  0,  &ReadInitZeroFlag     },
|    { "-n",  0,  &ReadNormalizeAllFlag },
|    { "-a",  0,  &ReadErrorWeightAlpha },
|    { "-b",  0,  &ReadErrorWeightBeta  },
|  };
|
|  int i, arg;
|  bool ret = TRUE;
|
|  Assert(argv);
|
|  for (i = 0; i < NUM_ELEMENTS(keys); i++)
|    keys[i].klen = strlen(keys[i].kname);
|
|  printf("Reading command-line overrides...\n");
|
[This should report an error if a match is not found in the list above.]
|  for (arg = 1; arg < argc; arg++)
|    for (i = 0; i < NUM_ELEMENTS(keys); i++)
|      if (strncmp(argv[arg], keys[i].kname, keys[i].klen) == 0)
|        ret &= keys[i].kread(argv[arg] + keys[i].klen, (FILE *) 0);
|
|  return ret;
|}



\Section{Skip Blanks / Non-Blanks}

   These routines are used in parsing the configuration input file.  The first
routine skips all blanks in a given routine, the second skips non-blanks.

\ENTRY: 's' points to a string.

\EXIT:  'SkipBlanks' and 'SkipNonBlanks' return a new pointer in the original
        string.

|static char *SkipBlanks (char *s)
|{
|  while (isspace(*s))
|    s++;
|
|  return s;
|}
|
|static char *SkipNonBlanks (char *s)
|{
|  while (*s && !isspace(*s))
|    s++;
|
|  return s;
|}



\Section{Read Input Line}

   This routine returns the next available source input line, discarding
comment lines as necessary.  The initial vertical bar and trailing newline
character are stripped here automatically.

\ENTRY: 's' points to a destination string.\n
        'slen' specifies the size of the destination string.\n
        'f' specifies an input file.

\EXIT:  's' contains a line of text (if the operation was successful).\n
        'ReadLine' returns 'TRUE' if successful or 'FALSE' at end-of-file.

|static bool ReadLine (char *s, int slen, FILE *f)
|{
|  while (fgets(s, slen, f))                 `Read until we receive a line
|  {                                         `beginning with a vertical bar.
|    if (s[0] == '|')
|    {
|      char *p;                              `Strip the trailing newline
|      if (p = strchr(s, '\n'))              `character and the vertical bar.
|        *p = '\0';
|      memmove(s, s+1, slen-1);
|      return TRUE;
|    }
|  }
|
|  return FALSE;
|}



\Section{Read Simple Components}

   These functions are called to read the number of input neurons, hidden-layer
neurons, and output neurons from the configuration file.  Range checking is
performed.

\ENTRY: 's' points to the relevant portion of an input line.

\EXIT:  All functions return 'TRUE' if successful, 'FALSE' otherwise.

|static bool ReadNumNeurons (char *s, FILE *f)
|{
|  c.nn = c.nh = atoi(s);
|  return CheckRange(c.nn, MIN_NEURONS, MAX_NEURONS,  "neurons");
|}
|
|static bool ReadNumInputs (char *s, FILE *f)
|{
|  c.nh -= (c.ni = atoi(s));
|  return CheckRange(c.ni, MIN_INPUT,  MAX_INPUT,  "inputs");
|}
|
|static bool ReadNumOutputs (char *s, FILE *f)
|{
|  c.nh -= (c.no = atoi(s));
|  return CheckRange(c.no, MIN_OUTPUT, MAX_OUTPUT, "outputs");
|}

The next group of functions are called to read other various values needing
range checking such as the the ploidy, the deme count, and so forth.

|static bool ReadPloidy (char *s, FILE *f)
|{
|  c.ploidy = atoi(s);
|  return CheckRange(c.ploidy, MIN_PLOIDY, MAX_PLOIDY, "ploidy");
|}
|
|static bool ReadNumDemes (char *s, FILE *f)
|{
|  c.ndemes = atoi(s);
|  return CheckRange(c.no, MIN_OUTPUT, MAX_OUTPUT, "demes");
|}
|
|static bool ReadNumOrganisms (char *s, FILE *f)
|{
|  c.npop = atoi(s);
|  return CheckRange(c.no, MIN_POP, MAX_POP, "organisms");
|}
|
|static bool ReadInitZeroFlag (char *s, FILE *f)
|{
|  c.initzero = atoi(s);
|  return CheckRange(c.initzero, FALSE, TRUE, "InitZero");
|}
|
|static bool ReadDiffEqFlag (char *s, FILE *f)
|{
|  c.diffeq = atoi(s);
|  return CheckRange(c.diffeq, FALSE, TRUE, "DiffEq Flag");
|}
|
|static bool ReadNormalizeAllFlag (char *s, FILE *f)
|{
|  c.normall = atoi(s);
|  return CheckRange(c.normall, FALSE, TRUE, "NormalizeAll Flag");
|}
|
|static bool ReadErrorWeightAlpha (char *s, FILE *f)
|{
|  c.walpha = (real) atof(s);
|  c.errorweight = TRUE;
|  return CheckRange(c.diffeq, FALSE, TRUE, "ErrorWeight Alpha");
|}
|
|static bool ReadErrorWeightBeta (char *s, FILE *f)
|{
|  c.wbeta = (real) atof(s);
|  c.errorweight = TRUE;
|  return CheckRange(c.diffeq, FALSE, TRUE, "ErrorWeight Beta");
|}

The remaining functions require no range checking.  That is, they have to
upper limit.

|static bool ReadRandSeed (char *s, FILE *f)
|{
|  c.seed = (uint32) atol(s);     `Seeds greater than $2^{31}$ are not supported.
|  return TRUE;
|}
|
|static bool ReadNumGenerations (char *s, FILE *f)
|{
|  c.ngen = atol(s);
|  return TRUE;
|}
|
|static bool ReadNumMutations (char *s, FILE *f)
|{
|  c.nmut = atoi(s);
|  return TRUE;
|}
|
|static bool ReadNumIterations (char *s, FILE *f)
|{
|  c.niter = atoi(s);
|  return TRUE;
|}



\Section{Read Variable Matrix}

   When the input file contains a variable matrix -- that is, a matrix
specifying exactly which elements are variable and which are constant -- the
matrix is slurped into memory.  '0' characters in the input file represent
constant 0 terms, '1' characters represent constant 1 terms, and all other
non-blank characters represent variable terms.  For example, the input

\S |Matrix
\S |  10000000000  0
\S |  01000000000  0
\S |  00100000000  0
\S |  00010000000  0
\S |  xxxx0000000  x
\S |  xxxx0000000  x
\S |  xxxx0000000  x
\S |  xxxx0000000  x
\S |  xxxx0000000  x
\S |  0000xxxxx00  x
\S |  0000xxxxx00  x

represents an 11-neuron system with 4 input neurons, 5 hidden-layer neurons,
and 2 output neurons.  The inputs remain constant, the hidden-layer and output
neurons may change.

   A general feedback mechanism is readily available by changing '0's to 'x's
in strategic locations.

[Perhaps this discussion really belongs at the front of this module...]

\ENTRY: 'c.M' and 'c.B' are assumed to be null.

\EXIT:  'ReadVarMatrix' returns 'TRUE' if successful, 'FALSE' otherwise.\n
        If successful, matrices 'c.M' and 'c.B' have been filled in.

|static bool ReadVarMatrix (char *s, FILE *f)
|{
|  int i, j;
|  char line[256], *p;
|
|  if (!CheckDefined(c.nn, "neurons"))
|    return FALSE;
|
|  c.M = MatCreate(c.nn, c.nn);
|  c.B = VecCreate(c.nn);
|
|  c.M->n = 0;
|  c.B->n = 0;
|
|  for (i = 0; i < c.nn; i++)
|  {
|    if (!ReadLine(line, sizeof(line), f))
|      return FALSE;
|
|    p = line;
|    for (j = 0; j < c.nn+1; j++)
|    {
|      #define  VAL(ch)  (real)(ch == '1'?  1 : 0)
|      #define  MUT(ch)  ((ch != '0') && (ch != '1'))
|
|      p = SkipBlanks(p);                    `Pluck one element from the input
|      if (!*p)                              `line.
|        return FALSE;
|
|      if (j < c.nn)                         `Store the element in the weight
|      {                                     `matrix or the threshold vector.
|        MatStore(c.M, i, j,
|                 VAL(*p), MUT(*p));
|      }
|      else
|      {
|        VecStore(c.B, i,
|                 VAL(*p), MUT(*p));
|      }
|
|      p++;
|    }
|  }
|
|  return TRUE;
|}



\Section{Load Default Variable Matrix}

   The default variable matrix allows for the hidden-layer and output neurons
to vary but fixes the input neurons at their original values.  In addition, the
default matrix uses no feedback, requiring only two iterations of the matrix.

\ENTRY: 'c.M' and 'c.B' are null.\n
        'c.nn', 'c.ni', 'c.nh', and 'c.no' must be defined.

\EXIT:  'LoadDefaultVarMatrix' returns void.\n
        Matrices 'c.M' and 'c.B' have been filled in.

|static void LoadDefaultVarMatrix (void)
|{
|  int i, j;
|
|  c.M = MatCreate(c.nn, c.nn);
|  c.B = VecCreate(c.nn);

Weight matrix.

|  c.M->n = 0;                               `Clear the entire weight matrix
|                                            `to zero.
|
|  for (i = 0; i < c.ni; i++)                `Fill the input section.
|    MatStore(c.M, i, i, (real)1, FALSE);
|
|  for (i = 0; i < c.nh; i++)                `Fill the hidden section.
|  for (j = 0; j < c.ni; j++)
|    MatStore(c.M, c.ni+i, j, (real)0, TRUE);
|
|  for (i = 0; i < c.no; i++)                `Fill the output section.
|  for (j = 0; j < c.nh; j++)
|    MatStore(c.M, c.ni+c.nh+i, c.ni+j, (real)0, TRUE);

Threshold vector.

|  c.B->n = 0;                               `Clear the entire threshold vector
|                                            `to zero.
|
|  for (i = 0; i < c.nh+c.no; i++)           `Fill the hidden and output
|    VecStore(c.B, c.ni+i, (real)0, TRUE);   `sections.
|}



\Section{Read Train/Test Set}

   These functions are called to read a training set or a test set from the
configuration file.  (C++ classes would be darn nice here.)

\ENTRY: 's' points to the relevant portion of an input line.
        'f' is ignored.

\EXIT:  Both functions return 'TRUE' if successful, 'FALSE' otherwise.

[These two routines should be folded together.]

|static bool ReadTrainSet (char *s, FILE *f)
|{
|  return ReadTrainTestSet(s, TRUE);
|}
|
|static bool ReadTestSet (char *s, FILE *f)
|{
|  return ReadTrainTestSet(s, FALSE);
|}

   This function does the actual work of reading the training and test sets;
the previous two functions merely call it with 'TRUE' and 'FALSE'.  (This was
broken apart from the two above functions as they had so much in common.  Speed
is not an issue here, so the repeated tests are not a problem.)

|static bool ReadTrainTestSet (char *s, bool train)
|{
|  int i;
|  Vector *X;
|
|  if (!CheckDefined(c.nn, "neurons") ||
|      !CheckDefined(c.ni, "inputs")  ||
|      !CheckDefined(c.nh, "hidden")  ||
|      !CheckDefined(c.no, "outputs"))
|    return FALSE;
|
|  if (!CheckMax(train?  c.ntrain        : c.ntest,
|                train?  MAX_TRAIN       : MAX_TEST,
|                train?  "training sets" : "test sets"))
|      return FALSE;
|
|  X = VecCreate(c.nn);                      `Create a fresh array.
|
|  for (i = 0; i < c.ni; i++)                `Read the input values.
|    X->m[0+i].r = ExtractReal(&s);
|
|  for (i = 0; i < c.no; i++)                `Read the output values (or set
|    X->m[c.ni+c.nh+i].r =                   `to zero in the test set case).
|      train? ExtractReal(&s) : (real) 0.0;
|
|  if (train)                                `Store the new array and advance
|    c.train[c.ntrain++] = X;                `the count.
|  else
|    c.test[c.ntest++] = X;
|
|  return TRUE;
|}



\Section{Extract Real Number from String}

   This routine extracts a single real number from a string and moves the
string pointer past the number, ignoring any white space encountered.

[This routine would be obsolete if I were to use 'strtok'...that would probably
be better?]

\ENTRY: 's' contains the address of a string pointer.

\EXIT:  'ExtractReal' returns a value.\n
        '*s' has been advanced past the value.

|static real ExtractReal (char **s)
|{
|  real r;
|  *s = SkipBlanks(*s);
|  r = (real) atof(*s);
|  *s = SkipNonBlanks(*s);
|  return r;
|}


\Section{Check Input Data}

This routine ensures that data values are neither too large nor too small.

\ENTRY: 'n' contains a value to check.\n
        'nmin' specifies the minimum allowable value.\n
        'nmax' specifies the maximum allowable value.\n
        'name' contains a description of the value, to be used in the case of
         a range error.

\EXIT:  'CheckRange' returns 'TRUE' if the value is conformant, 'FALSE'
        otherwise.

|static bool CheckRange (int n, int nmin, int nmax, char *name)
|{
|  if ((n >= nmin) && (n <= nmax))
|  {
|    return TRUE;
|  }
|  else
|  {
|    printf("Number of %s must be in the range %d to %d.\n",
|            name, nmin, nmax);
|    return FALSE;
|  }
|}


This routine ensures that data values are not too large.

\ENTRY: 'n' contains a value to check.\n
        'nmax' specifies the maximum allowable value.\n
        'name' contains a description of the value, to be used in the case of
         a range error.

\EXIT:  'CheckMax' returns 'TRUE' if the value is conformant, 'FALSE' otherwise.

|static bool CheckMax (int n, int nmax, char *name)
|{
|  if (n <= nmax)
|  {
|    return TRUE;
|  }
|  else
|  {
|    printf("Maximum number of %s is %d.\n", name, nmax);
|    return FALSE;
|  }
|}


   This routine ensures that a given value is defined.  When the program
starts, some values are set to the special value 'UNDEFINED'.  The value is
tested here to see if it has yet been defined.  Undefined values generate an
error message.

\ENTRY: 'n' contains a value to check.\n
        'name' contains a description of the value, to be used if the value
         is undefined.

\EXIT:  'CheckDefined' returns 'TRUE' if the value is defined, 'FALSE'
         otherwise.

|static bool CheckDefined (int n, char *name)
|{
|  if (n != UNDEFINED)
|  {
|    return TRUE;
|  }
|  else
|  {
|    printf("Missing number of %s in data file.\n");
|    return FALSE;
|  }
|}



\Section{Display Configuration}

   After a configuration has been loaded, it can be displayed to the standard
output device.  This is mainly a debugging device, and will not normally be
used.

\ENTRY: 'c' specifies a configuration set.

\EXIT:  The data has been written to 'stdout'.

|static void DisplayMatVar (Matrix *M, int i, int j)
|{
|  int n;
|
|  for (n = 0; n < M->n; n++)
|  {
|    if ((M->m[n].i == i) &&
|        (M->m[n].j == j))
|    {
|      if (M->m[n].mut)
|        printf("x");
|      else if (M->m[n].r == (real)0)
|        printf("0");
|      else
|        printf("1");
|      return;
|    }
|  }
|  printf("0");
|}
|
|static void DisplayVecVar (Vector *B, int i)
|{
|  int n;
|
|  for (n = 0; n < B->n; n++)
|  {
|    if (B->m[n].i == i)
|    {
|      if (B->m[n].mut)
|        printf("x");
|      else if (B->m[n].r == (real)0)
|        printf("0");
|      else
|        printf("1");
|      return;
|    }
|  }
|  printf("0");
|}

|void DisplayConfiguration (void)
|{
|  int i, j, t;

Display simple information.

|  printf("\n");
|  printf("CONFIGURATION\n");
|  printf("\n");
|  printf("Random number seed:  %ld\n", c.seed);
|  printf("\n");
|  printf("Total number of neurons:   %d\n", c.nn);
|  printf("Number of input neurons:   %d\n", c.ni);
|  printf("Number of hidden neurons:  %d\n", c.nh);
|  printf("Number of output neurons:  %d\n", c.no);
|  printf("\n");
|  printf("Number of demes (populations):          %d\n", c.ndemes);
|  printf("Number of organisms (population size):  %d\n", c.npop);
|  printf("\n");
|  printf("Maximum number of generations:  %ld %s\n", c.ngen,
|    c.ngen == 0?  "(no limit)" : "");
|  printf("\n");
|  printf("Number of mutations:   %d %s\n", c.nmut,
|    c.nmut == 0?  "(any number)" : "");
|  printf("Number of iterations:  %d\n", c.niter);
|  printf("\n");
|  printf("Organism initialization:  %s\n", c.initzero?  "Zero" : "Random");
|  printf("DiffEq flag:  %s\n", c.diffeq?  "True" : "False");
|  printf("NormalizeAll flag:  %s\n", c.normall?  "True" : "False");
|
|  if (c.errorweight)
|    printf("Synaptic Penalty:  f(x) = %f x^%f\n", c.walpha, c.wbeta);
|  else
|    printf("Synaptic Penalty:  None\n");
|  printf("\n");

Display variable matrix and locations.

|  printf("Variable matrix:\n");
|  for (i = 0; i < c.nn; i++)
|  {
|    printf("  ");
|    for (j = 0; j < c.nn; j++)
|      DisplayMatVar(c.M, i, j);
|    printf("  ");
|    DisplayVecVar(c.B, i);
|    printf("\n");
|  }
|  printf("\n");

Display training sets.

|  printf("%d Training sets:\n", c.ntrain);
|  for (t = 0; t < c.ntrain; t++)
|  {
|    VecDisplayRange(c.train[t], 0,         c.ni, "   Input", HDISPLAY);
|    VecDisplayRange(c.train[t], c.ni+c.nh, c.no, "  Output", HDISPLAY);
|    printf("\n");
|  }
|  printf("\n");

Display test sets.

|  printf("%d Test sets:\n", c.ntest);
|  for (t = 0; t < c.ntest; t++)
|  {
|    VecDisplayRange(c.test[t],  0,         c.ni, "   Input", HDISPLAY);
|    printf("\n");
|  }
|  printf("\n");
|}



TODD S. LEHMAN, JANUARY 1993

