\Module{Neural Network}

This is a program for research on artificial neural nets---specifically for
biological applications---though the program is general and has other uses.  The
neural net has a \E{dynamic structure} comprising the instantaneous state of
each neuron.  This dynamic structure is simply a set of real numbers, one number
representing the \E{activity level} of each neuron in the net.  The dynamic
structure changes with time.  The neural net also has a \E{static structure}
comprising three things for each neuron:
(1)~a threshold, also a real number,
(2)~a set of real coefficients representing weighted stimuli arriving from other
    neurons in the net, and
(3)~a set of real coefficients representing weighted responses sent to other
    neurons in the net.
The static structure does not change with time---or at it least changes on a
much slower time scale.  Changing the static structure is called \E{training}.

Some of the neurons are designated \E{input neurons}; their activity levels are
simply the values of the external stimuli attached to them.  Others are
designated \E{output neurons}; their activity levels are taken as responses of
the entire neural net and connected to an external object or to another neural
net.  The remainder are called hidden neurons.

Among the $n$ neurons in a network are at most $n^2$ synapses and $n$
thresholds.  Synapses are represented by a matrix $M$ of weight values,
thresholds by a vector $b$, and the state of the system at time $t$ by a vector
$x_t$:
$$
M   = \pmatrix{ w_{11} & w_{12} & \ldots & w_{1n} \cr
                w_{21} & w_{22} & \ldots & w_{2n} \cr
                \vdots & \vdots & \ddots & \vdots \cr
                w_{n1} & w_{n2} & \ldots & w_{nn} \cr }
\,\,\,\,\,\,
b   = \pmatrix{ b_1    \cr
                b_2    \cr
                \vdots \cr
                b_n    \cr }
\,\,\,\,\,\,
x_t = \pmatrix{ x_{t,1} \cr
                x_{t,2} \cr
                \vdots  \cr
                x_{t,n} \cr }
\,.
$$

In these terms, the neural net becomes nothing more than the discrete dynamical
system defined by
$$ x_{t+1} = F(Mx_t + b) \,, $$
where
$$ F(x) = \pmatrix{ f(x_1) \cr
                    f(x_2) \cr
                    \vdots  \cr
                    f(x_n) \cr } $$
and $f(u)=\tanh(u/T)$ for hidden neurons and $f(u)=u$ for input and output
neurons.  $T$ is a simulated 'temperature'---the degree of randomness in the
net.  The lower $T$, the sharper the non-linearity in $F$.  [Figure 1]

\midinsert
\input figtanh
{\narrower\narrower\noindent
%
% This needs to be a point or two smaller than the body of the document.
%
{\bf Figure 1.} The hyperbolic tangent function $f(u)=\tanh(u/T)$ for three
different values of $T$.  With $T=0$, the neural net becomes a digital device.

}
\endinsert

From this definition, $w_{ij}$ is the strength of the synapse by which neuron
$j$ stimulates neuron $i$.  The neural net is an iterated nonlinear map
on an $n$-dimensional real space $F:R^n \rightarrow R^n$.  All synchronous
neural nets that work from a sum of products and a threshold are special cases
of this model.

For example, consider a standard feed-forward neural net with four input
neurons, two hidden layers of three neurons each, and two output neurons.  This
is twelve neurons with only a small fraction of the $12\times12$ synapses
connected.  With neurons 1--4 as the inputs and neurons 11--12 as output, the
matrix has the structure.

\S                       1  2  3  4  5  6  7  8  9 10 11 12
\S                    1  1  -  -  -  -  -  -  -  -  -  -  -
\S                    2  -  1  -  -  -  -  -  -  -  -  -  -
\S                    3  -  -  1  -  -  -  -  -  -  -  -  -
\S                    4  -  -  -  1  -  -  -  -  -  -  -  -
\S                    5  X  X  X  X  -  -  -  -  -  -  -  -
\S                    6  X  X  X  X  -  -  -  -  -  -  -  -
\S                    7  X  X  X  X  -  -  -  -  -  -  -  -
\S                    8  -  -  -  -  X  X  X  -  -  -  -  -
\S                    9  -  -  -  -  X  X  X  -  -  -  -  -
\S                   10  -  -  -  -  X  X  X  -  -  -  -  -
\S                   11  -  -  -  -  -  -  -  X  X  X  -  -
\S                   12  -  -  -  -  -  -  -  X  X  X  -  -

Here 'X's represent real synapses (some value $w_{ij}\ne0$ takes the place of
each 'X'), dashes ('-') represent nonexistant synapses (where $w_{ij}=0$), and
'1's are used to hold values on input neurons.

As a second example, consider a Hopfield net. [Now describe this kind of
symmetric, fully connected network.]  [Also describe a competitive network.]

\H{Training}

[Include a section on genetic search explaining that the neural nets are trained
by simulated natural selection.  The program maintains a population of diploid
matrices.  Each cell of the operational matrix is the arithmetic mean (or
perhaps the sum) of the two haploid parts of the diploid matrix.  New matrices
are generated by recombination between  pairs of diploid matrices and by
mutation---using assortment among rows of the matrix and crossover within the
rows.  The genetic algorithm is independent of the rest of the program.]

\H{Format of the program}

The program is written in ANSI C in a document format.  Processed by a
\C{STAGE2} filter (Waite 1969) and further processed by the typesetting system
\TeX\ (Knuth 1986), it forms a document that describes the program in all
detail.  Processed by the utility 'd2d' it forms a program in normal C.  In the
makefile that accompanies this program, the command 'make program' creates an
executable version of the program and 'make document' creates a printable
document.  To prepare the document, one needs 'tex' and 'stage2'; to make the
program, one needs an \C{ansi} C compiler.  The document version is included in
the files labelled '*.dvi', and a version in a 300-dpi format is included in
files labelled '*.hp'.  To print a formatted copy of the document on a PC,
simply type the command 'copy /b *.hp lpt1'.

[We're not actually using \C{STAGE2}---we're using 'd2c'.]

[This last section does not belong here or in the makefile, but in an abstract
or something before the top-level table of contents.]

|#include <stdlib.h>
|#include <stdio.h>
-#include <conio.h>
|#include <string.h>
|#include <ctype.h>
|
|#include "neural.h"
|#include "protos.h"



\Section{Main}

This is the interface between the program and the operating system---in other
words the main entry point.  Command-line switches are defined in the
configuration module.  A file name is expected as the first non-switch
argument.

\ENTRY: 'argc' specifies the number of arguments to the program.\n
        'argv' is a list of pointers to the arguments.

\EXIT:  'main' returns a result code.\n
           0  No problems were encountered.\n
           1  Problems were encountered with the command line.\n
           2  A memory allocation failure occurred.

|#define  MAX_ORGANISMS  100
|
|int main (int argc, char *argv[])
|{
|  Org *u, *v, **(*pop);
|  char *configfile;
|  long gen;
|
|  int i, j, t;

|  printf("________________________________________"
|         "_______________________________________\n\n");

\H{Scan the command line for the configuration filename.}

|  configfile = (char *) 0;
|
|  for (i = 1; i < argc; i++)
|    if (argv[i][0] != '-')
|      configfile = argv[1];
|
|  if (!configfile)
|  {
|    printf("usage: neural [options] inputfile\n");
|    return 1;
|  }

\H{Read the configuration file.}

|  if (!ReadConfiguration(configfile) ||
|      !ReadArgConfiguration(argc, argv))
|    return 1;
|
|  DisplayConfiguration();
|
|  RandStart(c.seed);

\H{Create the initial population.}

|  pop = (Org ***) CreateArray2(c.ndemes, c.npop, sizeof(Org *));
|  for (i = 0; i < c.ndemes; i++)
|  for (j = 0; j < c.npop;   j++)
|  {
|    Org *o = OrgCreate(c.ploidy);
|    OrgInit(o);
|
|    if (c.initzero)
|      OrgZero(o);
|    else
|      OrgRand(o);
|
|    OrgEval(o);
|
|    pop[i][j] = o;                        `** I think I want OrgInsert here.
|  }

\H{Evolve the population until the user signals termination of the program.}

|  #define  KEY_NULL   0
|  #define  KEY_ESC    27
|  #define  KEY_SPACE  32
|
|  for (gen = 0; (!c.ngen || (gen < c.ngen)); gen++)
|  {
-    if (kbhit())                          `Check the keyboard and read a
|    if (0)                                `Check the keyboard and read a
|    {                                     `character if one is availble
|      int key = getch();                  `(MS-DOS) specific).
|
|      if (key == KEY_NULL)                `If the special null code is read,
|      {                                   `throw away the next character
|        getch();                          `without translation.  (MS-DOS)
|      }
|      else if (key == KEY_ESC)            `If the Escape key is pressed, halt.
|      {
|        break;
|      }
|      else if (key == KEY_SPACE)          `If the space bar is pressed,
|      {                                   `display the current weight and
|        MatDisplay(pop[0][0]->ploid[0].M, "M");` threshold matrices.
|        VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
-                                   ^ Temporary hack
|      }
|    }
|
|    for (i = 0; i < c.ndemes; i++)
|    {
|      u = pop[i][RAND_INT_RANGE(0, c.npop - 1)];
|      v = OrgClone(u);
|      OrgMutate(v);
|      OrgEval(v);
|
|      if (OrgInsert(pop[i], c.npop, v) == 0)
|      {
|        if (c.errorweight)
|          printf("Gen:%9ld      "
|                 "e-dist:%10.6f     w-dist:%10.6f     dist:%10.6f\n",
|                 gen, (double)v->edist, (double)v->wdist, (double)v->dist);
|        else
|          printf("Generation: %9ld            "
|                 "e-dist: %13.6f\n",
|                 gen, (double)v->edist);
|
-        OrgHolisticTest(v);
|      }
|
|      if (gen % 5000 == 0)
|      {
|        if (c.errorweight)
|          printf("Gen:%9ld\n", gen);
|        else                                   `<<-- yuck, clean this up!
|          printf("Generation: %9ld\n", gen);
|
|        MatDisplay(v->ploid[0].M, "M");` threshold matrices.
|        VecDisplay(v->ploid[0].B, "B", VDISPLAY);
|      }
|    }
|  }

Report final matrix.

|  MatDisplay(pop[0][0]->ploid[0].M, "M");
|  VecDisplay(pop[0][0]->ploid[0].B, "B", VDISPLAY);
-                              ^ Temporary hack

\H{Evalute test cases and display results}

|  u = pop[0][0];
-  ^^^^^^^^^^^^^ Temporary hack
|
|  OrgTest(u);
|
|  printf("Test results:\n");
|  for (t = 0; t < c.ntest; t++)
|  {
|    int i;
|
|    printf("   Input:");
|    for (i = 0; i < c.ni; i++)
|      printf("  %+6.3f", (double) c.test[t]->m[i].r);
|
|    printf("   Output:");
|    for (i = 0; i < c.no; i++)
|      printf("  %+8.3f", (double) c.test[t]->m[c.ni+c.nh+i].r);
|
|    printf("\n");
|  }
|  printf("\n");

\H{SPECIAL:  Test iteration of logistic equation}

|  if ((c.ni == 1) && (c.no == 1))
|  {
|    real x = 0.5;
-    real x = 0.152074;
|    long i = 0;
|
|    printf("SPECIAL ITERATION TEST:\n\n");
|
|    for (i = 0; i < 1000+100; i++)
|    {
-      if (kbhit())
|      if (0)
|      {
|        int key = getch();
|        if (key == KEY_NULL)
|          getch();
|        else if (key == KEY_ESC)
|          break;
|      }
|
|      if (i >= 1000)
|        printf("%6ld. %11.6f\n", i, x);
|
|      x = OrgIter(u, x);
|    }
|  }

\H{Destroy test organisms and exit}

|  for (i = 0; i < c.ndemes; i++)
|  for (j = 0; j < c.npop;   j++)
|    OrgDestroy(pop[i][j]);
|  DestroyArray2(pop);
|  return 0;
|}



TODD S. LEHMAN, JANUARY 1993.




