\Module{Random Number Generation}

    Here is a short collection of functions to generate random numbers.  These
functions may be used when standard library functions for random numbers are not
available or when precise characteristics of the numbers are important.  With
these functions, the random sequences should not vary with the machine, with the
operating system, or with the compiler.  The collection comprises four
functions:

\input table
$$
\BeginTable
\BeginFormat | l | l | \EndFormat
  \_
  | \T{Rand}         | Generate random number on the unit interval.    | \\+20
  \_
  | \T{RandStart}    | Start the sequence at a specified point.        | \\
  \_
  | \T{RandStartArb} | Start the sequence at an arbitrary point.       | \\
  \_
  | \T{RandInteger}  | Generate a random integer (the inner function). | \\+02
  \_
\EndTable
$$

    The first function in the collection is the most common in scientific
programming.  It merely generates a series of pseudo-random numbers on the
interval $0 \le X < 1$.  If it is the only function called, it will
generate the same sequence of numbers each time the program is run, thus
providing repeatability in numerical experiments.  A simple program that uses
it to list ten random numbers looks like this and prints the list that follows.

\S  main()
\S  {
\S    int n;
\S
\S    for (n = 0; n < 10; n++)
\S      printf("%f\n", Rand());
\S  }
\S
\S  0.211325
\S  0.544479
\S  0.220742
\S  0.111617
\S  0.893342
\S  0.290086
\S  0.212657
\S  0.105951
\S  0.686732
\S  0.749347



|#include <time.h>
|
|#include "neural.h"
|#include "protos.h"


\Section{Return random number in unit interval}

\ENTRY: No significant conditions.

\EXIT:  'Rand' contains the next random number in the sequence, uniformly
         distributed with $0 \le$ 'Rand' $< 1$.  Only the first 32 bits
         are significant; the remainder are zero (if the machine's arithmetic
         hardware is accurate).

|static unsigned long seed;

|double Rand (void)
|{
|  /* return((double)RandInteger() / 4294967296.); */
|  return((double)(RandInteger() >> 1) / 2147483648.);
|}


\Section{Initialize to a known starting point}

    By default, the function 'Rand' starts at the beginning of a long
deterministic sequence, continues for 4,294,967,296 steps, then repeats.
'Rand' can be started at any point within the sequence using the function
'RandStart'.  The function is passed a ``seed," which is an integer in the
range 0 to 4,294,967,295.  The default seed is zero.  With a seed of one, the
following program produces a different sequence.

\S  main()
\S  {
\S    int n;
\S
\S    RandStart((unsigned long) 1);
\S    for (n = 0; n < 10; n++)
\S      printf("%f\n", Rand());
\S  }
\S
\S  0.215868
\S  0.177158
\S  0.910775
\S  0.598857
\S  0.739466
\S  0.119943
\S  0.829061
\S  0.617727
\S  0.337687
\S  0.408679

\ENTRY: 'k' contains a starting random number seed.

\EXIT:  The random sequence is initialized to 'k'.

|unsigned long RandStart (unsigned long k)
|{
|  return(seed = k);
|}


\Section{Initialize to an arbitrary starting point}

    An arbitrary starting point is taken from the time of day.  Function
'RandStartArb' sets the random number seed to a value that will change from
time to time, so the sequence will likely start at a different point each time
its program is run (if enough time elapses).  The function returns the
beginning seed, which can be used with function 'RandStart' to repeat the
sequence.

\S  main()
\S  {
\S    unsigned long start;
\S    int n;
\S
\S    start = RandStartArb();
\S    for (n = 0; n < 10; n++)
\S      printf("%f\n", Rand());
\S
\S    printf("\n");
\S
\S    RandStart(start);
\S    for (n = 0; n < 10; n++)
\S      printf("%f\n", Rand());
\S  }
\S
\S  0.302036
\S  0.718284
\S  0.199497
\S  0.233959
\S  0.267500
\S  0.025543
\S  0.451980
\S  0.901616
\S  0.583625
\S  0.535184
\S
\S  0.302036
\S  0.718284
\S  0.199497
\S  0.233959
\S  0.267500
\S  0.025543
\S  0.451980
\S  0.901616
\S  0.583625
\S  0.535184

\ENTRY: No significant conditions.

\EXIT:  The random number sequence has been set to an arbitrary point.
        'RandStartArb' returns the random number seed, which can be
         used to restart the sequence.

|static unsigned long reverse (unsigned long k);

|unsigned long RandStartArb (void)
|{
|  static unsigned long base = 1234567;
|
|  base = base*5 + 1;
|  return(seed = base + reverse((unsigned long) time((long*)0)));
|}


\Section{Generate integer sequence}

    The actual random numbers are generated here as integers of 32 bits each,
created according to a linear congruential scheme.  Each successive number $x'$
in the series is generated from the preceding number $x$ by the relation
$x'=ax+c\pmod m$.  The multiplier $a$ is chosen so that $0<a-\sqrt m<m$, so that
$a\equiv5\pmod8$, and so that the arrangement of its bits appears haphazard.
The number $c$ is an odd integer as close to $(1-1/\sqrt3)\,m/2$ as possible.
The number $m$ is the length of the random number series to be generated, and is
taken here to be $2^{32}$.  These choices are sufficient to assure that the
numbers generated will reasonably imitate a random series.  See Donald E. Knuth,
\P{The Art of Computer Programming}, Volume 2 (Addison Wesley, 1969) for a
careful description of the method.

    The integer values can be used directly.  When they are, values less than
32 bits are usually needed, so only a portion of each generated number is used.
The portion taken should consist of a string of consecutive bits from the same
place in each number.  If a string of bits either from the high order or from
the central part of the number is taken, the resulting series will simulate
random numbers selected with replacement.  That is, duplicates and runs of the
same number can be expected.  If a string starting at the low order bit is
taken, the resulting series will simulate random numbers selected without
replacement.  In other words, each possible number will be generated once and
only once (then the series will repeat).

    On most machines, ``unsigned long" integers are precisely 32 bits in
length.  If they are longer than 32 bits, the constant 'OVERSIZE' should be
defined so that the operation '& 0xFFFFFFFF' is added to the code.  If they are
shorter than 32 bits, the function will not work. [**Update this discussion for
'limits.h'.]

\ENTRY: Static 'seed' contains the previous number in the sequence.

\EXIT:  'RandInteger' returns a new random number.\n
        'seed' contains the new random number for use in the next call.

|unsigned long RandInteger (void)
|{
|  #if ULONG_MAX > 0xFFFFFFFF
|    seed = (seed*19513957 + 907633385) & 0xFFFFFFFF;
|  #else
|    seed =  seed*19513957 + 907633385;
|  #endif
|
|  return seed;
|}


\Section{Reverse order of bits}

To convert slightly different times to largely different values, this function
reverses the order of bits in an integer.  Values of at least 32 bits are
assumed.

\ENTRY: 'k' contains a value to be reversed.

\EXIT:  'reverse' contains the value on entry with all bits in the opposite
         order.  That is, bit 0 becomes bit 31, bit 1 becomes bit 30, and so
         forth.

|static unsigned long reverse (unsigned long k)
|{
|  unsigned long h;
|  int n;
|
|  for (h = n = 0; n < 32; n++)
|  {
|    h = (h<<1) + (k&1);
|    k >>= 1;
|  }
|  return(h);
|}


CLARENCE L. LEHMAN, JULY 1972.

[Originally implemented on the IBM System 360]

{\parindent=2em
\item{1.}  Converted to 8080 language, May 1976 [CLL].
\item{2.}  Converted to document format, December 1987 [CLL].
\item{3.}  Converted to C, April 1988 [CLL].
\item{4.}  Expanded for UNIX, June 1991 [CLL].
\item{5.}  Converted to ANSI C, December 1992 [TSL].

}

This code has been placed in the public domain by its author.  If you
distribute this, please include all the comments.  If you must make any
changes, please note them above with your name or initials.

