\Module{Extract Source Code from Document}

This program processes a file written in document form and extracts the
source code, which is then intended for a normal compiler.  This module itself
is an example of the type of file processed.

To achieve a good turn of speed, the program interprets finite state tables,
where all the logic to extract the source code lies.  As currently written, five
states are used with 256 possible character inputs per state.  For each
character read, the two-dimensional 6\x256 state table is interrogated, with
state number indexing the row and character code indexing the column.  The state
table contains three action bits describing what is to be done with the
character and three state bits defining the next state.  The action bits tell
whether the character is to be written as output, whether it is to be counted as
a placeholder (blank), or whether accumulated blanks are to be written as
output.  These three actions together with state transitions suffice to extract
source code and discard trailing blanks.


|#define MSTATE  7                             `Binary mask for state bits.
|#define NSTATE  6                             `Maximum number of states.
|#define NCHAR   256                           `Size of character set.
|
|#define C 128                                 `Increment counter n.
|#define R  64                                 `Reset counter n.
|#define D  32                                 `Display n blanks.
|#define W  16                                 `Write current character.
|
|#define IN     getchar()                      `Character input.
|#define OUT(X) putchar(X)                     `Character output.
|
|static char tab[NSTATE][NCHAR];               `State transition tables.
|static char wmsg[] =
|  "/* THIS IS A DERIVED FILE -- DO NOT EDIT THIS VERSION */";


|main()
| {
| int state, currchar, value, n;               `Character counter.

| for (n=0; wmsg[n]; n++) OUT(wmsg[n]);        `Write a warning message.
|
| init();                                      `Initialize the state tables.
| n = state = 0;
|
| while (1)                                    `Read the next input character
| { currchar = IN;                             `and check for end of file.
|   if (currchar<=0) break;
|
|   value = tab[state][currchar];              `Locate the proper table entry
|   if (value & C) n += 1;                     `for the current state and
|   if (value & D) while (n-- > 0) OUT(' ');   `carry out any actions specified
|   if (value & R) n = 0;                      `for it.
|   if (value & W) OUT(currchar);
|
|   state = value & MSTATE;                    `Set the new state and repeat
|   }                                          `for the next input character.
|
| for (n=0; wmsg[n]; n++) OUT(wmsg[n]);        `Write a final message.
| OUT('\n');
| return(0);
| }

\Section{Initialize State Transition Tables}

This routine fills the state transition tables with logic to extract source
code from a document.  The meanings of the various states are described below.

|static char eol[]   = { '\n', 0 };            `End of line.
|static char null[]  = { '\r', 0 };            `Carriage return.
|static char bar[]   = { 124, 143, 0 };        `Beginning of source line.
|static char barc[]  = {  96, 143, 0 };        `End of source line.
|static char blank[] = { ' ', 128, 160, 0 };   `Blanks and pseudoblanks.
|static char other[] = { 0 };                  `All other characters.

|init()
| {

| set(tab[0], other, 1);                       `State 0.  At the beginning of a
| set(tab[0], null,  0);                       `new line.
| set(tab[0], eol,   0+W);
| set(tab[0], bar,   2);
|
| set(tab[1], other, 1);                       `State 1.  Within a full comment
| set(tab[1], eol,   0+W);                     `line.
|
| set(tab[2], other, 3+D+R+W);                 `State 2.  Blanks at beginning of
| set(tab[2], null,  2);                       `source line.
| set(tab[2], blank, 2+C);
| set(tab[2], eol,   0+R+W);
| set(tab[2], barc,  1+R);
|
| set(tab[3], other, 3+W);                     `State 3.  Within source line.
| set(tab[3], null,  3);
| set(tab[3], eol,   0+W);
| set(tab[3], blank, 4+C);
| set(tab[3], barc,  5);
|
| set(tab[4], other, 3+D+R+W);                 `State 4.  Blanks within source
| set(tab[4], null,  4);                       `line.
| set(tab[4], blank, 4+C);
| set(tab[4], eol,   0+R+W);
| set(tab[4], barc,  5+R);
|
| set(tab[5], other, 5);                       `State 5.  Within a comment to
| set(tab[5], eol,   0+W);                     `the right of source code.
| }

\Section{Set Values in Table}

The state tables must be initialized so that every character position
defines a new state and an optional set of actions to be performed before the
new state is entered.  This routine places values in the table for a specified
state.  It is passed a null-terminated string of characters together with
the action bits and new state they are to be assigned.  It sets all positions in
the table corresponding to characters in the string.  If, however, the string of
character is empty, then all positions in the table are set.

\ENTRY: 'itab' contains a table of state transitions constructed thus far.
         This table initially contains all zeros. \n
        'charlist' contains a list of characters to receive state transition
         values.  The list is terminated by a zero.  If the list is empty,
         then all characters not yet processed are to receive state transition
         values. \n
        'values' contains the state transition value.

\EXIT:  'itab' has been updated as described above.

|set(itab, charlist, value)
|
| char itab[], charlist[];
| int  value;
| {
| int i;
| unsigned char c;

| if (itab[0]==0)                              `If no characters were passed,
|   for (i=0; i<NCHAR; i++)                    `set the entire table.
|     itab[i] = value;
|
| else                                         `Otherwise set only the table
|   for (i=0; c=charlist[i]; i++)              `entries corresponding to the
|     itab[c] = value;                         `characters passed.
| }


\bigskip
\leftline{CLARENCE LEHMAN, DECEMBER 1989}
\medskip

\leftline{Modifications}
\medskip

\item{1.}  Vertical bar ('|') and opening single quote ('`') added to tables,
April 1992 [CLL].

\item{2.}  Converted to {\TeX} format, September 1992 [CLL].

\item{3.}  Output blank lines for full-line comments so that line numbers match
across source and derived files, December 1992 [TSL].

\item{4.}  Final new-line code added to end of file, April 2006 [CLL].

