package NeuralNet::Organism;

sub new ($$)   # constructor
{
	my ($class, $ploidy) = @_;
	$ploidy > 0 or die;

	my $nn = $NeuralNet::Config::nn;

	my $self = bless
	{
		ploidy => $ploidy,

		ploid  =>
		[
			map
			{
				{
					M => new NeuralNet::Matrix($nn, $nn),
					B => new NeuralNet::Matrix($nn, 1),
					X => new NeuralNet::Matrix($nn, 1),
				}
			}
			(0 .. $self->{ploidy}-1)
		],

		dist  => undef,
		edist => undef,
		wdist => undef,
	},
	$class;

	return $self;
}

sub clone ($)   # constructor
{
	my ($self) = @_;

	my $nn = $NeuralNet::Config::nn;

	my $clone = bless
	{
		ploidy => $self->{ploidy},

		ploid  =>
		[
			map
			{
				{
					M => $self->{ploid}[$_]{M}->clone(),
					B => $self->{ploid}[$_]{B}->clone(),
					X => $self->{ploid}[$_]{X}->clone(),
				}
			}
			(0 .. $self->{ploidy}-1)
		],

		dist  => $self->{dist},
		edist => $self->{edist},
		wdist => $self->{wdist},
	},
	ref($self);

	return $clone;
}

sub init ($)
{
	my ($self) = @_;

	foreach my $p (@{$self->{ploidy}})
	{
		$p->{M} = $NeuralNet::Config::M->clone();
		$p->{B} = $NeuralNet::Config::B->clone();
	}

	$self->{edist} = $self->{wdist} = $self->{dist} = 0;

	return;
}

sub initrand ($)
{
	my ($self) = @_;

	foreach my $p (@{$self->{ploidy}})
	{
		$p->{M}->randomize();
		$p->{B}->randomize();
	}

	return;
}

sub initzero ($)
{
	my ($self) = @_;

	foreach my $p (@{$self->{ploidy}})
	{
		$p->{M}->initzero();
		$p->{B}->initzero();
	}

	return;
}

sub mutate ($)
{
	my ($self) = @_;

	foreach my $p (@{$self->{ploidy}})
	{
		$p->{M}->mutate(0.1, 1.0);
		$p->{B}->mutate(0.1, 1.0);
	}

	return;
}

sub eval ($)
{
	my ($self) = @_;

	my $error = 0;
	foreach my $train (@NeuralNet::Config::train)
	{
		my $d = $self->train($train);
		$error += $d * $d;
	}
	$self->{edist} = sqrt($error / scalar(@NeuralNet::Config::train));

	$self->{wdist} = 0;
	if ($NeuralNet::Config::errorweight)
	{
		my $walpha = $NeuralNet::Config::walpha;
		my $wbeta  = $NeuralNet::Config::wbeta;

		foreach my $p (@{$self->{ploidy}})
		{
			$self->{wdist} +=
				$p->{M}->weight($walpha, $wbeta) +
				$p->{B}->weight($walpha, $wbeta);
		}

		$self->{wdist} /=
			$NeuralNet::Config::nn *
			($NeuralNet::Config::nn + 1) *
			$self->{ploidy};
	}

	return $self->{dist} = $self->{edist} + $self->{wdist};
}

sub train ($$)
{
	self ($self, $train) = @_;

	my $n = $NeuralNet::Config::ni + $NeuralNet::Config::nh;

	# [*** PLOIDY NOT IMPLEMENTED IN THIS ROUTINE ***]
	my $ox = $self->{ploid}[0]{X};

	my $i;
	for ($i = 0; $i < $NeuralNet::Config::ni; $i++)
	{
		$ox->{mat}{$i}{0} = $train->{mat}{$i}{0};
	}
	for (; $i < $NeuralNet::Config::nn; $i++)
	{
		$ox->{mat}{$i}{0} = 0;
	}

	$self->exec(0);   # 0 is a temporary hack

	if ($NeuralNet::Config::no == 1)
	{
		return abs($ox->{mat}{$n}{0} - $train->{mat}{$n}{0});
	}
	else
	{
		my $error = 0;
		for ($i = 0; $i < $NeuralNet::Config::no; $i++)
		{
			my $dist = $ox->{mat}{$n+$i}{0} - $train->{mat}{$n+$i}{0};
			$error += $dist * $dist;
		}
		return sqrt($error / $NeuralNet::Config::no);
	}
}

sub test ($)
{
	my ($self) = @_;

	my $n = $NeuralNet::Config::ni + $NeuralNet::Config::nh;

	# [*** PLOIDY NOT IMPLEMENTED IN THIS ROUTINE ***]
	my $ox = $self->{ploid}[0]{X};

	foreach my $test (@NeuralNet::Config::test)
	{
		my $i;
		for ($i = 0; $i < $NeuralNet::Config::ni; $i++)
		{
			$ox->{mat}{$i}{0} = $test->{mat}{$i}{0};
		}
		for (; $i < $NeuralNet::Config::nn; $i++)
		{
			$ox->{mat}{$i}{0} = 0;
		}

		$self->exec(0);   # 0 is a temporary hack

		for ($i = 0; $i < $NeuralNet::Config::no; $i++)
		{
			$test->{mat}{$n+$i}{0} = $ox->{mat}{$n+$i}{0};
		}
	}

	return;
}

sub exec ($$)
{
	my ($self, $ploidindex) = @_;

	my $p = $self->{ploid}[$ploidindex];
	$p or die;

	for (my $i = 1; $i <= $NeuralNet::Config::niter; $i++)
	{
		my $Y = NeuralNet::Matrix::product($p->{M}, $p->{X});
		$p->{X} = NeuralNet::Matrix::sum($p->{Y}, $p->{B});
		$self->norm();
	}

	return;
}

sub insert ($$)
{
	my ($self, $pop) = @_;
	ref $pop eq 'ARRAY' or die;

	my $i;

	for ($i = 0; $i < scalar(@$pop); $i++)
	{
		last if $self->{dist} < $pop->[$i]->{dist};
	}

	splice @$pop, $i, 0, $self;
	pop @$pop;

	return $i;
}

sub norm ($)
{
	my ($self) = @_;

	foreach my $p (@{$self->{ploid}})
	{
		if ($NeuralNet::Config::normall)
		{
			for (my $i = 0; $i < $NeuralNet::Config::nn; $i++)
			{
				$p->{X}{mat}{$i}{0} = normalize($p->{X}{mat}{$i}{0});
			}
		}
		else
		{
			my $ni = $NeuralNet::Config::ni;
			for (my $i = 0; $i < $NeuralNet::Config::nh; $i++)
			{
				$p->{X}{mat}{$ni+$i}{0} = normalize($p->{X}{mat}{$ni+$i}{0});
			}
		}
	}

	return;
}

sub normalize ($)   # not a method
{
	my ($x) = @_;

	my $T = 1.0;

	my $y = ($T == 0.0)
		? (($x >= 0.0)?  1 : -1)
		: exp(-2 * $x / $T);

	$y = (1 - $y) / (1 + $y);

	return ($y + 1) / 2;
}

1;
