#!/usr/bin/perl

# A script to turn documented C format (.d files) into stylized C
# files.  The documented C format is as follows: 
#
# - The file is written with TeX syntax, where all source code is
#   prefixed with a "|".  Comments not in the code region are in pure
#   TeX.
# - Comments within the code are prefixed with a "`" and end with a
#   ".".
# - Commented out lines of code are prefixed with a "-".
# - Lines that include notes about the previous line of code that are
#   intended to be only temporary are prefixed with a "\S" in some
#   cases.
#
# This script converts these to a stylized C format, which is as
# follows:
#
# - Comments start with the string "/*", followed by 76 dashes,
#   followed by a "*".  These comments may or may not be formatted
#   with TeX.
# - Comments end with the string "*/" on its own line.
# - Comments within the code are prefixed with "\\ ".
# - Commented out lines of code are prefixed with "\\".
#
# This script does not deal with the "\S" command when within code.
# These must be changed manually.  The "\S" command also appears to
# denote preformatted lines within the major comment bodies.  This
# syntax only appears in a subset of .d files, and thus, this should
# not be too much of a problem.

if ($#ARGV != 0) {
    print "\nUsage: ./d2sc.pl [FILE]\nFILE is the Documented C file to be converted";
}

open(FILE, $ARGV[0]);

$file = join("", <FILE>);
close(FILE);

# $file =~ s/([\n]{2,})([^|]*)([\n]{2,})/\1\/\/ \2\3/g;
$file =~ s/\n([|-][^\n]*)([\n]{2,})([^|\n])/\n\1\2\/\*----------------------------------------------------------------------------\*\n\3/g; # Opening comments
#$file =~ s/([-]*\*\\n.*)([\n]*[|-])/\1\n\*\/\2/sgm;
$file =~ s/(^[^|-][^\n]+)([\n]{1,})[|]/\1\n\*\/\2/sgm; # Closing comments
$file =~ s/[|](.*)/\1/g; # Removing "|" on code blocks
$file =~ s/\n-(.*)/\n\/\/\1/g; # Changing "-" lines to comments
# $file =~ s/\\S[^\S]/\/\//g; # The \S's were manually changed so they
                              # didn't interfere with the "real"
                              # documentation using the \S command
$file =~ s/(.*)`(.*)/\1 \/\/ \2/g; # Inline Comments
if (index($file, "/*") > index($file, "*/")) {
    $file = "\/\*----------------------------------------------------------------------------\*\n" . $file;
}
if (rindex($file, "/*") > rindex($file, "*/")) {
    $file .= "\*\/\n";
}

print $file;
