/*----------------------------------------------------------------------------*
\Module{Gradient descent}

This program will attempt to calculate the best network through
gradient descent.  In other words, it will start with a
randomly-selected point, and calculate all partial derivatives at that
point using a Taylor approximation.  Then, it will move a value of _H
in that direction.  It will repeat this until it has reached a local
minimum.

 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "neural.h"
#include "protos.h"

#define _H 0.01

// Function prototypes
void displayOrgInfo(real coords[], int ncoords);
real orgFitness(real coords[], int ncoords);

int main(int argc, char *argv[]) {
  int i,j;
  Org* o;
  // Load the config file
  char *configfile;
  configfile = (char *) 0;
  
  for (i = 1; i <  argc; i++)
    if (argv[i][0] != '-')
      configfile = argv[1];

  if (!configfile)
  {
    printf("usage: concavity [options] inputfile\n");
    return 1;
  }

  if (!ReadConfiguration(configfile) |
      !ReadArgConfiguration(argc, argv))
    return 1;

  DisplayConfiguration();

  // Initialize the random seed
  RandStart(c.seed);

  // Figure out how many mutable synapses were present in the config file
  int nsynapses = 0;
  for (i=0; i<c.M->n; i++)
    if (c.M->m[i].mut == TRUE)
      nsynapses++;
  for (i=0; i<c.B->n; i++)
    if (c.B->m[i].mut == TRUE)
      nsynapses++;

  // Read the input from stdin, stopping at the newline character or EOF
  #define MAXLINE 1024*1024 // No more than 1024 chars on a line
  char c;
  char *ch;
  char buff[MAXLINE];
  real coordinates[nsynapses];
  // Continue looping until we have reached the end of the stream.
  while (c != EOF) {
    for (i=0; i<MAXLINE; i++) {
      c = getc(stdin); // Using getc instead of gets to avoid a buffer overflow
      if (c == EOF || c == '\n') {
        buff[i] = '\0';
        break;
      }
      if (i == MAXLINE-1)
        printf("Error, input line too long.\n"), exit(5);
      buff[i] = c;
    }
    // Exit the program successfully on an empty line.
    if (strlen(buff) == 0) return 0;


    // We assume here that the number of columns in the input data is
    // equal to the number of synapses.  All coordinates from the
    // input line are split by the "," character and stored in
    // `coordinates`.
    ch = strtok(buff, ",");
    for (i=0; i<nsynapses; i++) {
      if (ch == NULL)
        printf("Error, insufficient coordinates.\n"), exit(7);
      coordinates[i] = atof(ch);
      ch = strtok(NULL, ",");
    }

    // This is how we know we have reached a max/min and can stop.
    // The highest fitness ever seen is stored in maxFitness.
    // Sometimes, a move in a direction of the gradient won't actually
    // increase fitness.  This effect is usually reversed, so
    // _TRIES_TO_MAX contains the number of times to keep trying after
    // having the current fitness not be higher than the maximum.
    // After this number of consecutive tries, the program terminates.
    real maxFitness = 0;
    #define _TRIES_TO_MAX 100
    int triesToMax = _TRIES_TO_MAX;

    // At each point, calculate the partial derivative in all
    // directions, and move in the direction of each partial
    // derivative.  Terminate as described above.
    while (1) {
        for (i=0; i<nsynapses; i++) {
            // Calculate fitness at the point
            real fi = orgFitness(coordinates, nsynapses);
            // Calculate the fitness + and - _H
            coordinates[i] += _H;
            real fiplus = orgFitness(coordinates, nsynapses);
            coordinates[i] -= 2*_H;
            real fiminus = orgFitness(coordinates, nsynapses);
            // Reset to the original value for the next iteration.
            coordinates[i] += _H;
            // First derivative
            real d1 = (fiplus-fiminus)/(2*_H);
            
            // This sets the conditions for whether we should increase
            // the value by 1, decrease the value by 1, or leave it
            // the same.
            if (d1 > .00001) {
                coordinates[i] -= _H;
            }
            else if (d1 < .00001) {
                coordinates[i] += _H;
            }
            
        }
        if (maxFitness >= orgFitness(coordinates, nsynapses)) {
            triesToMax--;
        }
        else {
            maxFitness = orgFitness(coordinates, nsynapses);
            triesToMax = _TRIES_TO_MAX;
        }
        if (triesToMax == 0)
            break;
    }
    printf("\n\nLOCAL MINIMUM FOUND!\n\n");
    displayOrgInfo(coordinates, nsynapses);
  }
  return 0;
}
/*
ENTRY CONDITIONS:
    `coords` is an array of M values in the order to be loaded
        directly into an organism, followed by a similar array of B
        values.  It must be `nsynapses` long.
EXIT CONDITIONS:
    Returns the fitness of the organism.
*/

real orgFitness(real coords[], int ncoords) {
  Org *o;
  real fitness;
  int i,j;
  o = OrgCreate(1);
  OrgInit(o);

  // We assume here for simplicity's sake that the coordinate inputs
  // are in the same order as in `o`.  (i.e. o->ploid[0].M->m[i].r
  // is the same as coords[i].)
  int k=0;
  for (i=0; i<o->ploid[0].M->n; i++) 
    if (o->ploid[0].M->m[i].mut)
      o->ploid[0].M->m[i].r = coords[k], k++;
  for (i=0; i<o->ploid[0].B->n; i++)
    if (o->ploid[0].B->m[i].mut)
      o->ploid[0].B->m[i].r = coords[k], k++;

  if (ncoords != k)
    printf("Error, check to make sure you are using the right config file.\n"), exit(8);
      
  // OrgEval takes us through all of the iterations necessary for
  // the organism, and puts the results in o->wdist, o->edist, and
  // o->dist.
  OrgEval(o);

  fitness = o->dist;
  OrgDestroy(o);
  return fitness;
}

// Displays information about the organism from a set of coordinates.
void displayOrgInfo(real coords[], int ncoords) {
  Org *o;
  real fitness;
  int i,j;
  o = OrgCreate(1);
  OrgInit(o);

  // We assume here for simplicity's sake that the coordinate inputs
  // are in the same order as in `o`.  (i.e. o->ploid[0].M->m[i].r
  // is the same as coords[i].)
  int k=0;
  for (i=0; i<o->ploid[0].M->n; i++) 
    if (o->ploid[0].M->m[i].mut)
      o->ploid[0].M->m[i].r = coords[k], k++;
  for (i=0; i<o->ploid[0].B->n; i++)
    if (o->ploid[0].B->m[i].mut)
      o->ploid[0].B->m[i].r = coords[k], k++;

  if (ncoords != k)
    printf("Error, check to make sure you are using the right config file.\n"), exit(8);
      
  // OrgEval takes us through all of the iterations necessary for
  // the organism, and puts the results in o->wdist, o->edist, and
  // o->dist.
  OrgEval(o);

  MatDisplay(o->ploid[0].M, "M");
  VecDisplay(o->ploid[0].B, "B", VDISPLAY);
  fitness = o->dist;
  printf("Final fitness is %f\n\n", fitness);

  printf("Concavity data:\n");
  for (i=0; i<ncoords; i++) 
    printf("%.6f,", coords[i]);
  printf("\n");

  OrgDestroy(o);
}

