/*----------------------------------------------------------------------------*
\Module{Matrix Operations}
*/

#include <stdio.h>
//#include <memory.h>
#include <math.h>
#include <string.h>

#include "neural.h"
#include "protos.h"



/*----------------------------------------------------------------------------*
\Section{Create Matrix/Vector}

[New discussion]

[Also explain how the matrix is "ready to go" when it leaves this routine.]

\ENTRY: 'di' specifies the height.\n
        'dj' specifies the width (not applicable in the vector case).

\EXIT:  'MatCreate' and 'VecCreate' return a pointer to the new matrix or
         vector.
*/

Matrix *MatCreate (int di, int dj)
{
  Matrix *A;
  int i, j, n;

  Assert(di > 0);
  Assert(dj > 0);

  A = (Matrix *) Alloc(sizeof(Matrix) + di*dj*sizeof(MatrixElement));

  n = 0;
  for (i = 0; i < di; i++)
  for (j = 0; j < dj; j++)
  {
    A->m[n].i = i;
    A->m[n].j = j;
    A->m[n].r = 0;
    A->m[n].mut = FALSE;
    n++;
  }

  A->di = di;
  A->dj = dj;
  A->n  = n;

  return A;
}

Vector *VecCreate (int di)
{
  Vector *A;
  int i, n;

  Assert(di > 0);

  A = (Vector *) Alloc(sizeof(Vector) + di*sizeof(VectorElement));

  n = 0;
  for (i = 0; i < di; i++)
  {
    A->m[n].i = i;
    A->m[n].r = 0;
    A->m[n].mut = FALSE;
    n++;
  }

  A->di = di;
  A->n  = n;

  return A;
}



/*----------------------------------------------------------------------------*
\Section{Destroy Matrix/Vector}

   This destroys a 2-dimensional matrix or a 1-dimensional vector.

\ENTRY: 'A' specifies a matrix or vector to destroy.

\EXIT:  'A' is undefined.
*/

void MatDestroy (Matrix *A)
{
  Assert(A);

  Release((void *) A);
}

void VecDestroy (Vector *A)
{
  Assert(A);

  Release((void *) A);
}



/*----------------------------------------------------------------------------*
\Section{Copy Matrix/Vector}

   These routines copiy the data of a source matrix or vector to a destination
matrix or vector of equal size.  No new storage is allocated; existing storage
is re-used.

\ENTRY: 'A' specifies a target matrix or vector.\n
        'B' specifies a source matrix or vector.

\EXIT:  'A' contains the contents of 'B'.\n
        'MatCopy' returns 'A'.
*/

Matrix *MatCopy (Matrix *A, Matrix *B)
{
  Assert(A->di == B->di);
  Assert(A->dj == B->dj);

  memcpy(A, B, sizeof(Matrix) + A->di*A->dj*sizeof(MatrixElement));

  return A;
}

Vector *VecCopy (Vector *A, Vector *B)
{
  Assert(A->di == B->di);

  memcpy(A, B, sizeof(Vector) + A->di*sizeof(VectorElement));

  return A;
}



/*----------------------------------------------------------------------------*
\Section{Store Matrix/Vector Value}

These routines store a single entry in a matrix or vector and increment the
element count.  Overflow is not checked unless assertion tracking is turned
on during compilation.

\ENTRY: 'A' specifies a matrix or vector.\n
        'i' contains the vertical coordinate of the element to store.\n
        'j' contains the horizontal coordinate of the element to store (not
         applicable in the vector case).\n
        'r' contains the value to store.\n
        'mut' specifies whether the value is mutable.
*/

void MatStore (Matrix *A, int i, int j, real r, bool mut)
{
  Assert(A);
  Assert(A->n < A->di*A->dj);

  A->m[A->n].i   = i;
  A->m[A->n].j   = j;
  A->m[A->n].r   = r;
  A->m[A->n].mut = mut;

  A->n++;
}

void VecStore (Vector *A, int i, real r, bool mut)
{
  Assert(A);
  Assert(A->n < A->di);

  A->m[A->n].i   = i;
  A->m[A->n].r   = r;
  A->m[A->n].mut = mut;

  A->n++;
}



/*----------------------------------------------------------------------------*
\Section{Mutate Matrix/Vector}

   These routines mutate a matrix or vector, perturbing it by random amounts
based on a probability of perturbation.  Nonmutable elements are not touched.

\ENTRY: 'A' specifies a matrix or vector.
        'p' specifies the probability of perturbation.
        'r' specifies a restriction factor on the perturbation.  Smaller
         values of 'r' result in generally smaller perturbations.

\EXIT:  'MatMutate' and 'VecMutate' return nothing.
*/

void MatMutate (Matrix *A, real p, real pz, real r)
{
  int n;

  Assert(A);
  bool mutated = FALSE;
  // Optimization: This while loop prevents the program through
  // everything if the matrix is still
  // exactly the same.
  while (mutated == FALSE) {
    for (n = 0; n < A->n; n++)
      if (A->m[n].mut && (Rand() < p)) {
        mutated = TRUE;
        if (Rand() < pz)
          A->m[n].r = 0;
        else
          A->m[n].r += RAND_REAL() * r;
      }
  }
}

void VecMutate (Vector *A, real p, real pz, real r)
{
  int n;

  Assert(A);

  for (n = 0; n < A->n; n++)
    if (A->m[n].mut && (Rand() < p)) {
      if (Rand() < pz)
        A->m[n].r = 0;
      else
        A->m[n].r += RAND_REAL() * r;
    }
}



/*----------------------------------------------------------------------------*
\Section{Compute Matrix/Vector Weight}

   These routines calculate the total synaptic weight of a matrix, adjusting
individual synapses along an exponential curve and scaling the final answer
linearly.

   The computed value is

   $\alpha {\Sum_{1\le i\le n}{1\le j\le n}} \abs{a_{ij}}^\beta}$,

where $\beta$ controls the exponential curve and $\alpha$ scales the total.

\ENTRY: 'A' specifies a matrix or vector.
        'alpha' scales the total.
        'beta' controls the exponential curve.
*/

real MatWeight (Matrix *A, real alpha, real beta)
{
  int n;
  real w;

  Assert(A);

  w = 0;
  for (n = 0; n < A->n; n++)
    w += pow(fabs(A->m[n].r), beta);

  return  alpha * w;
  }

real VecWeight (Vector *A, real alpha, real beta)
{
  int n;
  real w;

  Assert(A);

  if (!c.costonb) return 0;

  w = 0;
  for (n = 0; n < A->n; n++)
    w += pow(fabs(A->m[n].r), beta);

  return  alpha * w;
  }



/*----------------------------------------------------------------------------*
\Section{Add Vectors}

This routine computes the sum C of two matrices A and B.
$$ C = A + B $$
All three vectors must be equal in size.

(In $C++$ it should be possible to make this work nicely by overloading the
addition operator.)

\ENTRY: 'C' specifies the target matrix.\n
        'A' specifies the first matrix.\n
        'B' specifies the second matrix.

\EXIT:  'C' contains a non-sparse sum.  'mut' fields of the elements are
         undefined.
*/

void VecAdd (Vector *C, Vector *A, Vector *B)
{
  int n;

  Assert(A && B && C);
  Assert(A->di == C->di);
  Assert(B->di == C->di);

  C->n = C->di;
  for (n = 0; n < C->n; n++)               // Clear the target vector.
  {
    C->m[n].i = n;
    C->m[n].r = 0;
  }

  for (n = 0; n < A->di; n++)              // Add the first vector.
    C->m[A->m[n].i].r += A->m[n].r;

  for (n = 0; n < B->di; n++)              // Add the second vector.
    C->m[B->m[n].i].r += B->m[n].r;
}



/*----------------------------------------------------------------------------*
\Section{Multiply Matrix by Vector}

This routine computes the specialized product C of a matrix A and a vector B.
$$ C = A \times B $$
Four criteria are checked before the multiplication is performed:

{\parindent=2em
\item{1.} The destination area for target C must not equal the source area $B$.
\item{2.} The height of $B$ must equal the width of $A$.
\item{3.} The height of $C$ must equal the height of $A$.
\item{4.} The width of $C$ must equal the width of $B$.  (Guaranteed since
          both $C$ and $B$ are vectors.
\item{5.} The height of $C$ must equal the height of $B$.
}

If any of these criteria are not met, the program will panic and exit.

(In $C++$ it should be possible to make this work nicely by overloading the
multiplication operator.)


\ENTRY: 'C' specifies the target matrix.\n
        'A' contains the first matrix.\n
        'B' contains the second matrix.

\EXIT:  'C' contains the non-sparse product.  'mut' fields of the elements are
         undefined.
*/

void MatVecMult (Vector *C, Matrix *A, Vector *B)
{
  int n;

  Assert(A && B && C);
  Assert(C != B);
  Assert(B->di == A->dj);
  Assert(C->di == A->di);
  Assert(C->di == B->di);

  C->n = C->di;
  for (n = 0; n < C->n; n++)               // Clear the target vector.
  {
    C->m[n].i = n;
    C->m[n].r = 0;
  }

  for (n = 0; n < A->n; n++)               // Multiply.
  {
    int i = A->m[n].i;
    int j = A->m[n].j;

    C->m[i].r += A->m[n].r * B->m[j].r;
  }
}


/*----------------------------------------------------------------------------*
\Section{Iterate Matrices --- DiffEq Version}

   This special routine calculates the map
$$ x \mapsto x + G(x) dt $$

[** Get a description from Dad.]

[** This routine is still kind of a kludge in the sense that it should be
filtered through a more general interface.]

[** Design and usage of this routine is non-orthoganal.  It must be worked
nicely back in to the rest of the program.]


\ENTRY: 'Y' specifies the target matrix.\n
        'M' contains the weight matrix.\n
        'X' contains the state matrix.

\EXIT:  'Y' contains the product, subject to the above criteria.

//static real DiffEqNorm (real a, real x)
//{
//       if  (a > 0)  return  x / ( 1/a + x);
//  else if  (a < 0)  return -x / (-1/a + x);
//  else              return  0;
//}

/*----------------------------------------------------------------------------*
//void MatDiffEq (Matrix *Y, Matrix *M, Matrix *X)
//{
//  real dt = (real) (1.0);
//  int i, j;
//
//  Assert(Y && M && X);
//
//  for (i = 0; i < DI(M); i++)          @@ THIS STILL USES THE OLD
//  for (j = 0; j < DJ(X); j++)          @@ MATRIX PARADIGM!
//  {
//    real s = 0;
//    int k;
//
//    for (k = 0; k < DJ(M); k++)
//      s += DiffEqNorm(M[i][k], X[k][j]);
//
//    Y[i][j] = X[i][j] + s * dt;
//  }
//}



/*----------------------------------------------------------------------------*
\Section{Display Matrix/Vector}

These routines dump an array to 'stdout' for diagnostic use.

\ENTRY: 'A' specifies a matrix or vector.\n
        'name' specifies a name for the matrix or vector.

\EXIT:  'A' has been written to 'stdout'.
*/

static char *fmt_elem[2] =
{
  " %+7.7f",   // HDISPLAY (0)
  " %+7.7f\n"   // VDISPLAY (1)
};

static char *fmt_name[2] =
{
  "%s:",        // HDISPLAY (0)
  "%s =\n",     // VDISPLAY (1)
};

void MatDisplay (Matrix *A, char *name)
{
  int i, j;

  Assert(A);
  Assert(name);

  printf(fmt_name[VDISPLAY], name);

  for (i = 0; i < A->di; i++)
  {
    for (j = 0; j < A->dj; j++)
    {
      int n;
      real r = 0;

      for (n = 0; n < A->n; n++)
      {
        if ((A->m[n].i == i) &&
            (A->m[n].j == j))
        {
          r = A->m[n].r;
          break;
        }
      }

      printf(fmt_elem[HDISPLAY], r);
    }

    printf("\n");
  }

  printf("\n");
}

void VecDisplayRange (Vector *A, int i0, int di, char *name, int format)
{
  int i;

  Assert(A);
  Assert(name);

  printf(fmt_name[format], name);

  for (i = i0; i < i0+di; i++)
  {
    int n;
    real r = 0;

    for (n = 0; n < A->n; n++)
    {
      if (A->m[n].i == i)
      {
        r = A->m[n].r;
        break;
      }
    }

    printf(fmt_elem[format], r);
  }

  if (format == VDISPLAY)
    printf("\n");
}

void VecDisplay (Vector *A, char *name, int format)
{
  Assert(A);
  VecDisplayRange(A, 0, A->di, name, format);
}

/*----------------------------------------------------------------------------*
\Section{Lookup Matrix/Vector Value}

These routines retrieve a value from a matrix/vector.

\ENTRY: 'A' specifies a matrix or vector.
        'i' is the row to lookup.
        'j' is the column to lookup (only for the matrix version).

\EXIT:  Returns the value located at at the position $M_{i,j}$ or $B_i$.
*/

real MatLookup (Matrix *M, int i, int j) {
    Assert(i < M->di);
    Assert(j < M->dj);
    int k;
    // Loop through to find the correct element
    for (k=0; k<M->n; k++)
        if (M->m[k].i == i && M->m[k].j == j)
            return M->m[k].r;
    return 0; // If it doesn't exist, it is zero.
}

real VecLookup (Vector *V, int i) {
    Assert(i < V->di);
    int k;
    // Loop through to find the correct element
    for (k=0; k<V->n; k++)
        if (V->m[k].i == i)
            return V->m[k].r;
    return 0; // If it doesn't exist, it is zero.
}

/*----------------------------------------------------------------------------*
\Section{Revision History}

1.  Original revision, December 1992 [TSL].
2.  Revised for new matrix paradigm, December 1993 [TSL].
3.  Added lookup functions, July 2012 [MWS].


*/
