/*----------------------------------------------------------------------------*
\Module{Neural Network}

[This header file is a conglomeration of several earlier header files.]



\Section{General Types}

[description]

   The ANSI C header file 'limits.h' is used here to decide which integer types
to use for various derived fixed-size aliases.
*/

#include <limits.h>


/*----------------------------------------------------------------------------*
Shorter, clearer aliases necessary to program in C.
*/

typedef  unsigned char   uchar;
typedef  unsigned short  ushort;
typedef  unsigned int    uint;
typedef  unsigned long   ulong;

typedef  unsigned int    bool;


/*----------------------------------------------------------------------------*
8-bit integer types.
*/

#if    (UCHAR_MAX == 0xFF)

  typedef  uchar   uint8;
  typedef   char    int8;

#else

  #error No definition for "uint8"

#endif


/*----------------------------------------------------------------------------*
16-bit integer types.
*/

#if    (UINT_MAX == 0xFFFF)

  typedef  uint    uint16;
  typedef   int     int16;

#elif  (USHRT_MAX == 0xFFFF)

  typedef  ushort  uint16;
  typedef   short   int16;

#else

  #error No definition for "uint16"

#endif


/*----------------------------------------------------------------------------*
32-bit integer types.
*/

#if    (UINT_MAX == 0xFFFFFFFF)

  typedef  uint    uint32;
  typedef   int     int32;

#elif  (ULONG_MAX == 0xFFFFFFFF)

  typedef  ulong   uint32;
  typedef   long    int32;

#else

  #error No definition for "uint32"

#endif


/*----------------------------------------------------------------------------*
Real number type.

 typedef float real;            WHY WON'T THIS WORK RIGHT???  CASTING PROBLEMS?
*/
typedef double real;


/*----------------------------------------------------------------------------*
Useful constants.
*/

#define  TRUE   1
#define  FALSE  0


/*----------------------------------------------------------------------------*
Useful macros.  [**describe**]
*/

#define  NUM_ELEMENTS(x)  (sizeof(x) / sizeof((x)[0]))



/*----------------------------------------------------------------------------*
\Section{Dynamic Arrays}

   The 'NEW' macro allocates a specified structure from the heap and may be
released with the 'Release' function.  Instead of writing this,

\S foo *p = (foo *) Alloc(sizeof(foo));

it is possible to write this:

\S foo *p = NEW(foo);

\ENTRY:  't' specifies a basic or derived type.

\EXIT:   'NEW(t)' is a pointer to a new object of type 't'.
*/

#define  NEW(t)  ((t *) Alloc(sizeof(t)))



/*----------------------------------------------------------------------------*
\Section{Matrices and Vectors}

[OLD PARADIGM]

   A matrix is simply a dynamically allocated two-dimensional array of real
numbers; a vector is simply a dynamically allocated one-dimensional array of
real numbers.  Please see the array module for more details.

//typedef  real  **Matrix;
//typedef  real   *Vector;

/*----------------------------------------------------------------------------*
[NEW PARADIGM]

   The new matrix paradigm is a one-dimensional array of matrix elements.
This array is capable of representing sparce matrices efficiently (credit to
R. McGehee for this data structure).  Each element contains a matrix value
$r$ and a matrix location $(i,j)$.  There is no particular order of the
elements.
*/

typedef struct
{
  int i;                     // Vertical coordinate.
  int j;                     // Horizontal coordinate.
  real r;                    // Value.
  bool mut;                  // Mutable if true.
}
MatrixElement;

typedef struct
{
  int di;                    // Height of 2D matrix.
  int dj;                    // Width of 2D matrix (set to 1 for vector).
  int n;                     // Current number of participating elements.
//  MatrixElement m[];         // di*dj slots for participating elements.
  MatrixElement m[1];        // di*dj slots for participating elements.
}
Matrix;

/*----------------------------------------------------------------------------*
   A vector is similar, but does not have a horizontal component.
*/

typedef struct
{
  int i;                     // Vertical coordinate.
  real r;                    // Value.
  bool mut;                  // Mutable if true.
}
VectorElement;

typedef struct
{
  int di;                    // Height of 1D matrix.
  int n;                     // Current number of participating elements.
//  VectorElement m[];         // di slots for participating elements.
  VectorElement m[1];        // di slots for participating elements.
}
Vector;



/*----------------------------------------------------------------------------*
\Section{Configuration}

   Variable matrix codes.  These are used to distinguish among constant zero,
constant one, and variable (e.g., mutatable) elements of the weight matrix and
threshold vectors.
*/

#define  VAR_MUTATE  (real)(-1.0)
#define  VAR_ZERO    (real)( 0.0)
#define  VAR_ONE     (real)( 1.0)

/*----------------------------------------------------------------------------*
   Matrix template codes.  These are similar to the variable matrix codes in
that they codify types of states, but are different in that they are arbitrary
integers.  For speed in switch statements, these numbers should be consecutive
starting with zero (so the compiler can do a table-lookup and jump).  However,
nothing will break if this is not the case.

@@ Will VAR_xxxx become obsolete?
*/

#define  TEMPLATE_0  0
#define  TEMPLATE_1  1
#define  TEMPLATE_X  2



/*----------------------------------------------------------------------------*
   Main configuration structure.  Here is held all of the data extracted from
the configuration file and from the command line.  Many of the substructures
are allocated dynamically at run-time based on neuron counts, population size,
and so forth.

   The model weight matrix 'M' and threshold vector 'B' are obtained from the
configuration file and are used in preparing new organisms.
*/

typedef struct
{
  ulong     seed;                  // Random number generator seed.

  int       nn;                    // Total number of neurons.
  int       ni;                    // Number of Input neurons.
  int       nh;                    // Number of Hidden-layer neurons.
  int       no;                    // Number of Output neurons.

  int       ploidy;                // Ploidy of organisms (1, 2, etc.)
  int       ndemes;                // Number of populations (demes).
  int       npop;                  // Number of organisms (population size).
  long      ngen;                  // Number of generations.
  real      pmut;                  // Probability of mutation on each synapse.
  real      pzero;                 // Probability of mutation on each synapse.
  int       niter;                 // Number of iterations for organism.

  Matrix    *M;                    // Weight matrix.
  Vector    *B;                    // Threshold vector.

  int       ntrain;                // Number and array of training sets.
  Vector    **train;

  int       ntest;                 // Number and array of test sets.
  Vector    **test;

  bool      initzero;              // Initialization mode (T = zero, F = rand).
  bool      diffeq;                // Differential equations matrix flag.
  bool      normall;               // Normalize all neurons flag.
  bool      normweights;           // Normalize weights in calculating w-dist
  bool      stepnormal;            // Step function flag.

  bool      errorweight;           // Consider neuron weight in error calculation.
  real      walpha;                // Weight alpha --- scalar multiplier.
  real      wbeta;                 // Weight beta --- exponent.
  bool      costonb;               // TRUE to include B when calculating alpha/beta cost.

  real      sigma;                 // If non-zero, add noise to training sets with stdv=sigma

  char      *outfile;              // The file to which data is written.
  char      *initcoords;           // Initial coordinates the starting population should be set to
}
Config;


/*----------------------------------------------------------------------------*
   The configuration structure is kept globally accessible so that it doesn't
need to be passed around to dozens of routines (previous versions of the
program passed a pointer in, and it got pretty ugly).
*/

extern Config c;



/*----------------------------------------------------------------------------*
\Section{Organism Operations}

   Ploid structure.  A haploid organism consists of a single set of weight,
state, and threshold matrices.  A diploid organism consists of a pair of sets.

[** Perhaps X belongs OUTSIDE this structure...........examine ramifications.]
*/

typedef struct
{
  Matrix *M;                 // Weight matrix.
  Vector *X;                 // State vector.
  Vector *B;                 // Threshold (boundary) vector.
}
Ploid;


/*----------------------------------------------------------------------------*
   Organism structure.  Haploid and diploid organisms are constructed by
setting the ploidy to 1 or 2, respectively, and allocating a corresponding
number of ploid structures (defined above).
*/

typedef struct _Org
{
  struct _Org  *link;        // Pointer to next organism in the chain.  Used
                             // only in the available list.

  int     ploidy;            // Ploidy of organism (1, 2, etc.)
  Ploid   *ploid;            // Poly-ploidic genetic makeup array.

  real    edist;             // Distance (error) from ideal output set.
  real    wdist;             // Weight distance from empty network.
  real    dist;              // Total distance (sum of distance components).
}
Org;


/*----------------------------------------------------------------------------*
   Aliases are created here for the haploid and diploid case.  The ploidy of an
organism is set to either HAPLOID (1) or DIPLOID (2).
*/

#define  HAPLOID   1         // Organism ploidy identifiers.
#define  DIPLOID   2


/*----------------------------------------------------------------------------*
   For the case of a diploid organism, aliases are created here for the
maternal and paternal portions.  The 'ploid' array in the organism structure
(defined above) is indexed by MATERNAL (0) or PATERNAL (1).
*/

#define  MATERNAL  0         // Array indexes for the two halves of a diploid
#define  PATERNAL  1         // organism.



/*----------------------------------------------------------------------------*
\Section{Error Handling}


Assert macro.  Can be enabled or disabled.  The cost is relatively low when
enabled, and zero when disabled.
*/

#if ASSERT

  #define  Assert(expr)  if (!(expr)) AssertFunc(__FILE__, __LINE__, #expr)

#else

  #define  Assert(expr)

#endif


/*----------------------------------------------------------------------------*
Checkpoint macro.  Used only a couple of times in debugging.  Will not appear
in any final version of the program.
*/

#define  CHECKPOINT()  printf("Checkpoint:  %s(%d)\n", __FILE__, __LINE__);



/*----------------------------------------------------------------------------*
\Section{Random Number Generation}

[**All of these would be nicer as functions. A bit slower, but a lot cleaner.]

   The 'RAND_INT_RANGE' macro produces random integers in a specified closed
interval $[min,max]$.
*/

#define  RAND_INT_RANGE(min, max)  (int) \
  (min + (RandInteger() >> 16) * (long)(max - min + 1) / 65536L)


/*----------------------------------------------------------------------------*
   The 'RAND_RANGE' macro produces random real numbers in a specified interval.
Since the function 'Rand' returns a number in the interval $[0,1)$, this macro
returns a number in the interval $[min,max)$.
*/

#define  RAND_RANGE(min, max)  (min + (Rand() * (max - min)))


/*----------------------------------------------------------------------------*
   The 'RAND_REAL' macro produces random numbers over the range of real
numbers, with a bias toward small numbers.  A random number is chosen in the
range $-\pi/2$ to $\pi/2$, then the tangent of that number is calculated.  In
order to avoid infinitely large numbers (that is, practically speaking, machine
overflow), we use the arctangent of a very large number rather than exactly
$\pi/2$.
*/

#define  ATAN_BIG      1.570796326695        //  $atan(10^{10})$
#define  RAND_REAL()   (real) tan(RAND_RANGE(-ATAN_BIG, ATAN_BIG))



/*----------------------------------------------------------------------------*
\Section{Formatting constants}

Arrays, vectors, and perhaps some other objects can be displayed either
horizontally or vertically.
*/

#define  HDISPLAY  0
#define  VDISPLAY  1


