/*----------------------------------------------------------------------------*
\Module{Organism Operations}
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "neural.h"
#include "protos.h"


static Org *avail = (Org *) 0;

#define MAX_LINELENGTH 1024*1024
#define MAX_SYNAPSES 100

/*----------------------------------------------------------------------------*
\Section{Create Organism}

   New organisms are created by allocating a new organism data structure and an
array of ploidy sets.  Each ploidy set consists of three matrices, so for a
diploid organism there are $1 + 1 + 2 \times 3 = 8$ calls to the general
storage allocator.

   Older organisms on the unused list are re-used here if the list is
non-empty.

\ENTRY: 'c' specifies a model configuration.
        'ploidy' specifies the ploidy of the organism (1, 2, etc.).

\EXIT:  'OrgCreate' returns a pointer to an organism.\n
        The ploid array and its matrices have been allocated.
*/

Org *OrgCreate (int ploidy)
{
  Org *o;
  int i;

  Assert(ploidy > 0);

  if (avail)                                 // If a previously used structure is
  {                                          // available, fetch it from the list
    o = avail;                               // of unused organisms.  Constituent
    avail = avail->link;                     // matrices are already in place.
  }

  else                                    
  {                                       
    o = (Org *) Alloc(sizeof(Org));          // Otherwise allocate a brand new
    o->ploidy = ploidy;                      // structure.  Create new weight,
    o->ploid = (Ploid *)                     // threshold, and state matrices.
               CreateArray1(o->ploidy,
                            sizeof(Ploid));   

    for (i = 0; i < o->ploidy; ++i)       
    {                                     
      Ploid *p = &o->ploid[i];               // Loop through all the ploids in
      p->M = MatCreate(c.nn, c.nn);          // the organism, allocating weight,
      p->B = VecCreate(c.nn);                // threshold, and state matrices for
      p->X = VecCreate(c.nn);                // each ploid.
    }

  }

  return o;
}



/*----------------------------------------------------------------------------*
\Section{Clone Organism}

   This routine clones an organism by creating a new one and copying all the
data of the original.  The clone is a new instance allocated dynamically.

\ENTRY: 'v' specifies a parent organism.

\EXIT:  'OrgClone' returns a pointer to a new but equivalent organism.
*/

Org *OrgClone (Org *o)
{
  Assert(o);

  return OrgCopy(OrgCreate(o->ploidy), o);
}



/*----------------------------------------------------------------------------*
\Section{Copy Organism}

   This routine copies all the data of a source organism into a destination
organism of equal size.

\ENTRY: 'u' specifies a destination organism.\n
        'v' specifies a source organism.

\EXIT:  'OrgCopy' returns 'u'.
*/

Org *OrgCopy (Org *u, Org *v)
{
  int n;

  Assert(u && u->ploid);
  Assert(v && v->ploid);
  Assert(u->ploidy == v->ploidy);

  for (n = 0; n < u->ploidy; ++n)            // Loop through all ploids.
  {
    Ploid *p = &u->ploid[n];
    Ploid *q = &v->ploid[n];

    Assert(p->M && p->B && p->X);
    Assert(q->M && q->B && q->X);

    MatCopy(p->M, q->M);                     // Copy the weight, threshold, and
    VecCopy(p->B, q->B);                     // state matrices.
    VecCopy(p->X, q->X);
  }

  u->edist = v->edist;                       // Copy the error.
  u->wdist = v->wdist;
  u->dist  = v->dist;

  return u;
}



/*----------------------------------------------------------------------------*
\Section{Destroy Organism}

   This routine returns an organism to a list of unused organisms.

\ENTRY: 'o' specifies an organism to destroy.

\EXIT:  'OrgDestroy' returns nothing.\n
        'o' is undefined.
*/

void OrgDestroy (Org *o)
{
  Assert(o && o->ploid);

  if (1)
  {
    o->link = avail;                         // Insert the organism at the head
    avail = o;                               // of the list of unused organisms.
  }

  else
  {
    int n;                                   // Destroy the organism.  (This is
    for (n = 0; n < o->ploidy; ++n)          // never actually used, but is
    {                                        // provided for completeness.)
      Ploid *p = &o->ploid[n];
      Assert(p->M && p->B && p->X);
      MatDestroy(p->M);                   
      VecDestroy(p->B);                   
      VecDestroy(p->X);                   
    }
    DestroyArray1(o->ploid);
    Release(o);
  }
}



/*----------------------------------------------------------------------------*
\Section{Initialize Model Organism}

   This routine takes the default variable matrix and copies it into an
organism's initial matrix as an initialization step.

\ENTRY: 'o' specifies an organism.\n
        'c' specifies a model configuration.

\EXIT:  'OrgInit' returns 'o'.
*/

Org *OrgInit (Org *o)
{
  int n;

  Assert(o && o->ploid);

  for (n = 0; n < o->ploidy; ++n)            // Loop though the ploids, copying
  {                                          // in the weight matrix and the
    Ploid *p = &o->ploid[n];                 // threshold vector.
    Assert(p->M && p->B);
    MatCopy(p->M, c.M);
    VecCopy(p->B, c.B);
  }

  o->edist = (real) 0.0;                     // Clear the error to zero.
  o->wdist = (real) 0.0;
  o->dist  = (real) 0.0;

  return o;
}



/*----------------------------------------------------------------------------*
\Section{Initialize Random Organism}

   This routine initializes an organisms's initial matrices to random values.
Some searches work better when not started at zero.  [**I'm not exactly sure
why yet--has something to do with local maxima...]

\ENTRY: 'o' specifies an organism.\n
        'c' specifies a model configuration.

\EXIT:  'OrgRand' returns 'o'.
*/

Org *OrgRand (Org *o)
{
  int q;

  Assert(o && o->ploid);

  for (q = 0; q < o->ploidy; ++q)            // Examine each ploid.
  {
    Ploid *p = &o->ploid[q];
    int n;

    Assert(p->M && p->B);

    for (n = 0; n < p->M->n; n++)            // Complete the weight matrix.
      if (p->M->m[n].mut)
        p->M->m[n].r = RAND_REAL();

    for (n = 0; n < p->B->n; n++)            // Complete the threshold vector.
      if (p->B->m[n].mut)
        p->B->m[n].r = RAND_REAL();
  }
  return o;
}



/*----------------------------------------------------------------------------*
\Section{Initialize Zero Organism}

   This routine zeros out variable locations in the weight matrix and threshold
array.  The internal representation after loading the configuration file is
that variable locations are marked with -1, so these are changed to 0's.

\ENTRY: 'o' specifies an organism.\n
        'c' specifies a model configuration.

\EXIT:  'OrgZero' returns 'o'.
*/

Org *OrgZero (Org *o)
{
  int q;

  Assert(o && o->ploid);

  for (q = 0; q < o->ploidy; ++q)            // Examine each ploid.
  {
    Ploid *p = &o->ploid[q];
    int n;

    Assert(p->M && p->B);

    for (n = 0; n < p->M->n; n++)            // Complete the weight matrix.
      if (p->M->m[n].mut)
        p->M->m[n].r = 0;

    for (n = 0; n < p->B->n; n++)            // Complete the threshold vector.
      if (p->B->m[n].mut)
        p->B->m[n].r = 0;
  }
  return o;
}


/*----------------------------------------------------------------------------*
\Section{Initialize Predifined Organism}

   This routine zeros out variable locations in the weight matrix and threshold
array.  The internal representation after loading the configuration file is
that variable locations are marked with -1, so these are changed to 0's.

\ENTRY: 'o' specifies a monoploid organism.\n
        'c' specifies a model configuration.

\EXIT:  'OrgZero' returns 'o'.
*/

Org *OrgFromCoords (Org *o, char icoords[]) {
  // If there is a default, load it.  This obviously only works for
  // 1-ploid organisms.
  int i;
  char *ch;
  char buff[MAX_LINELENGTH];
  real coordinates[MAX_SYNAPSES];
  char *icoordscopy = strdup(icoords);
  // Continue looping until we have reached the end of the string
  for (i=0; i<MAX_LINELENGTH; i++) {
    if (icoordscopy[i] == '\0') {
      icoordscopy[i+1] = '\0';
      break;
    }
    if (i == MAX_LINELENGTH-1)
      printf("Error, initial coordinates input line too long.\n"), exit(5);
  }

  ch = strtok(icoordscopy, ",");
  for (i=0; i<MAX_SYNAPSES; i++) {
    if (ch == NULL)
      break;
    coordinates[i] = atof(ch);
    ch = strtok(NULL, ",");
  }
  int k=0;
  for (i=0; i<o->ploid[0].M->n; i++) 
    if (o->ploid[0].M->m[i].mut)
      o->ploid[0].M->m[i].r = coordinates[k], k++;
  for (i=0; i<o->ploid[0].B->n; i++)
    if (o->ploid[0].B->m[i].mut)
      o->ploid[0].B->m[i].r = coordinates[k], k++;

  MatDisplay(o->ploid[0].M, "Name");

  return o;
}  

/*----------------------------------------------------------------------------*
\Section{Mutate Organism}

   This routine mutates one or more values in the weight matrix and/or the
threshold vector of an organism.  All ploids are affected.

\ENTRY: 'o' specifies an organism.

\EXIT:  The organism has been mutated.
*/

void OrgMutate (Org *o)
{
  int p;

  Assert(o && o->ploid);

  for (p = 0; p < o->ploidy; p++)
  {
    MatMutate(o->ploid[p].M, c.pmut, c.pzero, 1.0);
    VecMutate(o->ploid[p].B, c.pmut, c.pzero, 1.0);
//                                            ^^^
  }
}



/*----------------------------------------------------------------------------*
\Section{Evaluate Organism}

   This routine computes the weakness of an organism -- that is, the difference
from a theoretical ` // perfect-fit'' organism:  the closer to zero, the better
the fit.  For a given a configuration with a set of training values, the
distance in n-dimensional space between the organism's output and the expected
output from the training values is computed and is assumed to be the weakness
of the organism.

\ENTRY: 'o' specifies an organism.\n
        'c' specifies a configuration for use with training values.

\EXIT:  'o->dist' contains the total weakness.\n
        'o->edist' contains the distance from the training sets.
        'o->wdist' optionally contains the synapse weight factor.
        'OrgEval' returns the value of 'o->dist'.
*/

real OrgEval (Org *o)
{
  real error;
  int i;

  Assert(o);

  error = (real) 0.0;
  for (i = 0; i < c.ntrain; i++)             // For each of the training sets,
  {                                          // compute the weighted Euclidean
    real d = OrgTrain(o, c.train[i]);        // distance to the expected (desired)
    error += d * d;                          // output set.
  }

  o->edist = (real) sqrt(error / c.ntrain);  // Calculate the weighted Euclidean
                                             // distance (root mean square) over
                                             // all training values.

  o->wdist = 0;                              // Optionally add in a factor which
  if (c.errorweight)                         // assigns a penalty to non-zero
  {                                          // synapse connections.
    int p;
    for (p = 0; p < c.ploidy; p++)
    {
      o->wdist += MatWeight(o->ploid[p].M, c.walpha, c.wbeta);
      o->wdist += VecWeight(o->ploid[p].B, c.walpha, c.wbeta);
    }
    if (c.normweights)
        o->wdist /= (real)(c.nn * (c.nn + 1) * c.ploidy);
  }

  return  o->dist = o->edist + o->wdist;     // Return final answer.
}



/*----------------------------------------------------------------------------*
\Section{Evaluate Organism for Particular Training Value}

   This routine computes the weakness of an organism for a particular training
value.  [**More documentation here...]

\ENTRY: 'o' specifies an organism.\n
        'train' specifies a training set.

\EXIT:  'OrgTrain' returns the weakness.

[*** PLOIDY NOT IMPLEMENTED IN THIS ROUTINE ***]
*/

real OrgTrain (Org *o, Vector *train)
{
  int n = c.ni + c.nh;
  int i;

  VectorElement *ox = o->ploid[0].X->m;
  VectorElement *tx = train->m;

  Assert(o && o->ploid);
  Assert(train);

  for (i = 0; i < c.ni; i++)                 // Initialize the state vector.
    ox[i].r = tx[i].r;
  for (; i < c.nn; i++)
    ox[i].r = 0;



  OrgExec(o, 0);                             // Execute the organism.  (Heh.)
//             ^ Temporary hack

  if (c.no == 1)                             // If there is only one output neuron,
  {                                          // simply return the difference in
    return (real)                            // values.
      fabs(ox[n].r - (Gaussian(c.sigma) + tx[n].r)); // Add noise
  }

  else                                       // If there is more than one output
  {                                          // neuron, compute the distance in
    real error = 0;                          // n-dimensional space.
    int i;

    for (i = 0; i < c.no; i++)
    {
      real dist = ox[n+i].r - (Gaussian(c.sigma) + tx[n+i].r); // Add noise
      error += dist * dist;
    }

    return (real) sqrt(error / c.no);        // Return the weighted (RMS) distance.
  }
}



/*----------------------------------------------------------------------------*
\Section{Test Organism}

[description]

[*** PLOIDY NOT IMPLEMENTED IN THIS ROUTINE ***]
*/

void OrgTest (Org *o)
{
  int n = c.ni + c.nh;
  int t;

  Assert(o && o->ploid);

  for (t = 0; t < c.ntest; t++)              // Loop through the test sets.
  {
    int i;

    VectorElement *ox = o->ploid[0].X->m;
//                                 ^ Temporary hack
    VectorElement *tx = c.test[t]->m;

    for (i = 0; i < c.ni; i++)               // Initialize the state vector.
      ox[i].r = tx[i].r;
    for (; i < c.nn; i++)
      ox[i].r = 0;

    OrgExec(o, 0);                           // Execute the organism.  (Heh.)
//               ^ Temporary hack

    for (i = 0; i < c.no; i++)               // Record the result.
      tx[n+i].r = ox[n+i].r;
  }
}



/*----------------------------------------------------------------------------*
\Section{Holistic Test}

[description]

@@ Ploidy is assumed to be 1 here!
*/

static real WriteHolistic (Org *o, real *r, real z, real h)
{
  real z1, dg;

  *r += h;
  z1 = OrgEval(o);
  *r -= h;

  dg = (z1 - z) / h;

  printf("%+7.3f", dg);

  return  dg * dg;
}

static void BlankHolistic (void)
{
  printf("    ---");
}

void OrgHolisticTest (Org *o)
{
  real g, z;
  int i, j, n;

  real h = (real) 0.001;

  Matrix *M = o->ploid[0].M;
  Vector *B = o->ploid[0].B;

  printf("\nHolistic test:\n");

  OrgEval(o);
  z = o->dist;

  g = 0;

  for (i = 0; i < M->di; i++)
  {
    for (j = 0; j < M->dj; j++)
    {
      for (n = 0; n < M->n; n++)
        if ((M->m[n].i == i) &&
            (M->m[n].j == j))
          g += WriteHolistic(o, &M->m[n].r, z, h);

      if (n == M->n)
        BlankHolistic();
    }

    printf("    ");

    for (n = 0; n < B->n; n++)
      if (B->m[n].i == i)
        g += WriteHolistic(o, &B->m[i].r, z, h);

    if (n == B->n)
      BlankHolistic();

    printf("\n");
  }

  printf("\nLength of gradient vector = %7.3f\n\n", sqrt(g));
}



/*----------------------------------------------------------------------------*
\Section{Execute Organism}

   This routine is called once for each ploid in an organism.  Execution
involves ` // executing'' the neural network -- that is, iterating the system of
matrices.  Each iteration step involves multiplying the weight matrix by the
state vector, then adding the threshold matrix, then normalizing the state
vector.

   A static work vector is permanently allocated for speed.

\ENTRY: 'o' specifies an organism.
        'p' specifies a ploid by index.

\EXIT:  'o->X' contains the result vector.
*/

void OrgExec (Org *o, int p)
{
  Matrix *M;
  Vector *B, *X;
  static Vector *Y = (Vector *) 0;

  int i;

  Assert(o && o->ploid);
  Assert(p < o->ploidy);

  M = o->ploid[p].M;
  B = o->ploid[p].B;
  X = o->ploid[p].X;

  if (!Y)                                    // Allocate the work vector only once.
    Y = VecCreate(X->di);

  for (i = 1; i <= c.niter; i++)
  {
    if (c.diffeq)
    {
 //      MatDiffEq(Y, M, X);                    // [DiffEq case.  Description, etc.]
 //      MatCopy(X, Y);
 //    The above two lines started with == in document form.
    }
    else
    {
      MatVecMult(Y, M, X);                   // Iterate the neural network,
      VecAdd(X, Y, B);                       // normalizing hidden-layer and
      OrgNorm(o);                            // output neurons between iterations.
    }
  }
}



/*----------------------------------------------------------------------------*
\Section{Iterate simple 1-input, 1-output network}

(This was created specially for the logistic equation test in late October '93.)

** NO PLOIDY SUPPORT HERE YET **

\ENTRY: 'o' points to an organism.
        'x' contains a single input value for the single input neuron.

\EXIT:  'OrgIter' returns the value of the single output neuron.
*/

real OrgIter (Org *o, real x)
{
  int i;

  Assert(c.ni == 1);
  Assert(c.no == 1);

  o->ploid[0].X->m[0].r = x;                 // Set the input value.
//           ^ Hack

  for (i = c.ni; i < c.nn; i++)              // Clear out the hidden neurons.
    o->ploid[0].X->m[i].r = 0;

  OrgExec(o, 0);                             // Execute the organism.  (Heh.)
//             ^ Temporary hack
  
  return o->ploid[0].X->m[c.nn-1].r;         // Return the output neuron.
}



/*----------------------------------------------------------------------------*
\Section{Insert Organism into Population}

   When a new organism is created it is added to one of the demes (populations)
and another organism is kicked out to make room.  The demes are ordered by
error (field 'dist' of the organism structure).

\ENTRY: 'pop' specifies a deme (a simple array of organisms).
        'npop' contains the size of the deme.
        'o' points to a new organism.

\EXIT:  One of two cases has occurred:
          a. either the new organism has been added to the deme and an older
             organism has been destroyed to make room, or
          b. the organism was is not worthy of inclusion in the deme and has
             simply been destroyed.

[** Needs to be smart about npop = 0 **]

[Should probably work as a linked list, not an array.  The link is already in
place in the 'Org' structure, so it won't be much of a change.]
*/

int OrgInsert (Org *pop[], int npop, Org *o)
{
  int i, n;

  Assert(pop);
  Assert(o);

  for (i = 0; i < npop; i++)               // Search for a spot to insert the new
    if (o->dist < pop[i]->dist)            // organism.  (This would be faster as
      break;                               // a binary search.)

  if (i < npop)                            // If the new organism is worthy of
  {                                        // keeping, destroy the organism at
    OrgDestroy(pop[npop-1]);               // the bottom of the list, shift the
                                           // list to make room for the new
    n = npop - i - 1;                      // new organism (by doing a 'memmove'
    if (n > 0)                             // on the array of pointers), and save
      memmove(&pop[i+1], &pop[i],          // the new organism in the new spot.
        n * sizeof(Org *));

    pop[i] = o;
  }

  else                                     // Otherwise don't even bother keeping
  {                                        // the organism; just destroy it
    OrgDestroy(o);                         // without touching the rest of the
  }                                        // population.

  return i;
}



/*----------------------------------------------------------------------------*
\Section{Normalize Organism Output Matrix}

The normalization function is hard-coded here.  It could just as well be a
member function in the organism data structure or a runtime configuration
choice.

\ENTRY: 'o' specifies an organism.

\EXIT:  The hidden layer of 'o->X' has been normalized.
*/

void OrgNorm (Org *o)
{
  int n;

  Assert(o && o->ploid);

  for (n = 0; n < o->ploidy; ++n)
  {
    Vector *X = o->ploid[n].X;
    int i;

    Assert(X);

    if (c.normall)                         // If this is a Boolean-type network,
    {                                      // normalize all the neurons.
      for (i = 0; i < c.nn; i++)
        X->m[i].r = Normalize(X->m[i].r);
    }
    else                                   // Otherwise normalize only the
    {                                      // hidden-layer neurons.
      for (i = 0; i < c.nh; i++)
        X->m[c.ni+i].r = Normalize(X->m[c.ni+i].r);
    }
  }
}



/*----------------------------------------------------------------------------*
\Section{Normalization Function}

This particular normalization function uses the formula
   $$f(x) = tanh(x/T) = {{1-e^{-2x/T}}\over{1+e^{-2x/T}}}$$
where $T$ represents a ` // temperature.''  The smaller the value of $T$, the
` // squarer'' the edge on the function.

The approximate cost of this function is 5 floating-point operations.

\ENTRY: 'x' contains an unnormalized value in the range $(-\infty,\infty)$.

\EXIT:  'Normalize' returns a value in the interval range $(-1,1)$.
*/

#define  T  1.0

real Normalize (real x)
{
  real y;

  if (T == 0.0 || c.stepnormal)            // Special case: handle 'T = 0' binary function so that
    y = (x >= 0.0)?  1 : 0;                // we don't divide by zero.

  else                                     // General case: divide by 'T' safely.
    y = (real) exp(-2 * (double) x / T);

  double temp = x;                         // Test for infinity, which creates a NaN error
  if ((temp == y) && ((temp - y) != 0.0))
    return -1;

//  return (1 - y) / (1 + y);  // @@ -1 to +1
  y = (1 - y) / (1 + y);     // @@  0 to +1
  return (y + 1) / 2;
}

/*----------------------------------------------------------------------------*
\Section{Inflict noise on the population}

This function looks through the population it was passed and applies
noise with the sigma specified in the configuration file.  This
function is necessary because in previous simulations involving noise,
it was clear that some organisms entered the population due to
particular noisy conditions, and since the fit was so good, it never
left, even though it does not fit well to other noisy conditions.
This will ensure that they fit well to any noisy function, which, in
turn, will probably cause the population to change hands rather quickly.

I don't know whether this is the best method to implement noise or
not.  With this method, all individuals in the population are
undergoing a different type of nosie.  I should try this both using
this method, and giving each particular generation a noise array to
apply to both the population and the newly-generated organism.

\ENTRY: 'pop' contains a pointer to a population of organisms

\EXIT: Applies different amounts of noise to every organism in the
       population, re-evaluates them all, and then sorts them again.

*/

void PopEvalNoise(Org* pop[])
{
  int i;
  // First evaluate them all while subjected to noise
  for (i=0; i<c.npop; i++) {
    OrgEval(pop[i]);
  }
  // Now sort the entire list based on their new fitness.  Since n is
  // small, the list is mostly sorted, and I'm feeling lazy, this
  // seems like a good time for Bubblesort.
  Org* u, v;
  bool sorted = FALSE;
  while (!sorted) {
    sorted = TRUE;
    for (i=0; i<c.npop-1; i++) {
      if (pop[i]->dist > pop[i+1]->dist) {
        u = pop[i];
        pop[i] = pop[i+1];
        pop[i+1] = u;
        sorted = FALSE;
      }
    }
  }
  
}

/*----------------------------------------------------------------------------*
FERMI FUNCTION:  y = 1 / exp(-2 * (double) x / T);  no (y+1)/2 deal
*/


/*----------------------------------------------------------------------------*
\Section{Generate a Gaussian Random variable}

This function uses the Box-Muller algorithm to generate a Gaussian
random variable around a mean of zero.  Since this algorithm by nature
generates two random variables, it would be more efficient to return
two values through pointers, though this would add complexity and is
probably not worth the additional computational time.

\ENTRY: 'sigma' contains the desired standard deviation

\EXIT:  Returns a random real from a Gaussian distribution around 0
*/

real Gaussian(double sigma) {
    if (sigma==0) return (real)0;
    double r1, r2;
    r1 = Rand(); r2 = Rand();   // 0 <= R1,R2 < 1
    // We do log(1-r1) instead of log(r1) so r1 < 0
    double n1 = sqrt(-2*log(1-r1))*cos(3.141592653589793*2*r2);
    return (real)n1*sigma;
}

/*
\Section{Revision History}

1.  Original revision, December 1992 [TSL].
2.  Revised for new matrix paradigm, December 1993 [TSL].

*/
