/*----------------------------------------------------------------------------*
\Module{Calculate partial derivatives at a point}

This program takes as its input (through stdin) lines of data points,
comma delimited.  It assumes that these data points are points at which 
all first derivatives are zero.  Data points must be formatted as:

M1,M2,M3,...,B1,B2,B3,...

The neuron numbers are those specified in the config file.  Run this
with the same config file that was used to run the simulation and
everything will work out correctly.  It then will print Octave code to
stdout that will plot all partial derivatives at a point.

Note that this steals a significant amount of code from "concavity.c".
I am also quite proud of the fact that this compiled and ran the first
time without any bugs.

 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "neural.h"
#include "protos.h"
#include "concavity.h"

// Specifies how far to go in each direction from the specified point
#define DISTANCE 110
// Specifies how big the step size should be.  This float should
// divide DISTANCE.
#define STEP .1

int main(int argc, char *argv[]) {
  int i,j;
  Org* o;
  // Load the config file
  char *configfile;
  configfile = (char *) 0;
  
  for (i = 1; i <  argc; i++)
    if (argv[i][0] != '-')
      configfile = argv[1];

  if (!configfile)
  {
    printf("usage: concavity [options] inputfile\n");
    return 1;
  }

  if (!ReadConfiguration(configfile) |
      !ReadArgConfiguration(argc, argv))
    return 1;

  // DisplayConfiguration();

  // Initialize the random seed
  RandStart(c.seed);

  // Figure out how many mutable synapses were present in the config file
  int nsynapses = 0;
  for (i=0; i<c.M->n; i++)
    if (c.M->m[i].mut == TRUE)
      nsynapses++;
  for (i=0; i<c.B->n; i++)
    if (c.B->m[i].mut == TRUE)
      nsynapses++;

  // Read the input from stdin, stopping at the newline character or EOF
  #define MAXLINE 1024*1024 // No more than 1024*1024 chars on a line
  char c;
  char *ch;
  char buff[MAXLINE];
  real coordinates[nsynapses];
  int plotcounter = 0; // Count plots when multiple arguments given
  // Continue looping until we have reached the end of the stream.
  while (c != EOF) {
    for (i=0; i<MAXLINE; i++) {
      c = getc(stdin); // Using getc instead of gets to avoid a buffer overflow
      if (c == EOF || c == '\n') {
        buff[i] = '\0';
        break;
      }
      if (i == MAXLINE-1)
        printf("Error, input line too long.\n"), exit(5);
      buff[i] = c;
    }
    // Exit the program successfully on an empty line.
    if (strlen(buff) == 0) return 0;


    // We assume here that the number of columns in the input data is
    // equal to the number of synapses.  All coordinates from the
    // input line are split by the "," character and stored in
    // `coordinates`.
    ch = strtok(buff, ",");
    for (i=0; i<nsynapses; i++) {
      if (ch == NULL)
        printf("Error, insufficient coordinates.\n"), exit(7);
      coordinates[i] = atof(ch);
      ch = strtok(NULL, ",");
    }

    printf("\n\n# BEGIN PLOT %i\n\nfigure\n", plotcounter);
    for (i=0; i<nsynapses; i++) {
      real tmpcoord = coordinates[i];
      coordinates[i] = -DISTANCE;
      printf("\n\n# COORDINATE %i:\n\n", i);
      printf("subplot(%i, 2, %i);\n", (nsynapses+1)/2, i+1);
      printf("A = [");
      while (coordinates[i] <= DISTANCE) {
        real fit = orgFitness(coordinates, nsynapses);
        printf("%.8f,%.8f\n", coordinates[i], fit);
        coordinates[i] += STEP;
      }
      coordinates[i] = tmpcoord;
      printf("];\n"
             "plot(A(:,1),A(:,2), [%f], [%f], \"*\", \"markersize\", 10);\n", tmpcoord, orgFitness(coordinates,nsynapses));
      printf("axis([-%i %i 0 1.2]);\n", DISTANCE, DISTANCE);
    }
    printf("\npause()\n\n"
           "# END PLOT %i\n\n", plotcounter);
    plotcounter++;

  }
  return 0;
}

/*
ENTRY CONDITIONS:
    `coords` is an array of M values in the order to be loaded
        directly into an organism, followed by a similar array of B
        values.  It must be `ncoords` long.
EXIT CONDITIONS:
    Returns the fitness of the organism.
*/

real orgFitness(real coords[], int ncoords) {
  Org *o;
  real fitness;
  int i,j;
  o = OrgCreate(1);
  OrgInit(o);

  // We assume here for simplicity's sake that the coordinate inputs
  // are in the same order as in `o`.  (i.e. o->ploid[0].M->m[i].r
  // is the same as coords[i].)
  int k=0;
  for (i=0; i<o->ploid[0].M->n; i++) 
    if (o->ploid[0].M->m[i].mut)
      o->ploid[0].M->m[i].r = coords[k], k++;
  for (i=0; i<o->ploid[0].B->n; i++)
    if (o->ploid[0].B->m[i].mut)
      o->ploid[0].B->m[i].r = coords[k], k++;

  if (ncoords != k)
    printf("Error, check to make sure you are using the right config file.\n"), exit(8);
      
  // OrgEval takes us through all of the iterations necessary for
  // the organism, and puts the results in o->wdist, o->edist, and
  // o->dist.
  OrgEval(o);

  fitness = o->dist;
  //printf("Fitness is %f", fitness);
  OrgDestroy(o);
  return fitness;
}

